import threading
import time
from queue import Queue

import numpy as np

from asrp.interface import WhisperInference, HFSpeechInference


def live_vad_process(device_name, asr_input_queue, vad_mode=2):
    import webrtcvad
    import pyaudio
    vad = webrtcvad.Vad()
    vad.set_mode(vad_mode)

    audio = pyaudio.PyAudio()
    FORMAT = pyaudio.paInt16
    CHANNELS = 1
    RATE = 16000
    # A frame must be either 10, 20, or 30 ms in duration for webrtcvad
    FRAME_DURATION = 30
    CHUNK = int(RATE * FRAME_DURATION / 1000)

    microphones = live_list_microphones(audio)
    selected_input_device_id = live_get_input_device_id(
        device_name, microphones)

    stream = audio.open(input_device_index=selected_input_device_id,
                        format=FORMAT,
                        channels=CHANNELS,
                        rate=RATE,
                        input=True,
                        frames_per_buffer=CHUNK)

    frames = b''
    while True:
        if LiveSpeech.exit_event.is_set():
            break
        frame = stream.read(CHUNK, exception_on_overflow=False)
        is_speech = vad.is_speech(frame, RATE)
        if is_speech:
            frames += frame
        else:
            if len(frames) > 1:
                asr_input_queue.put(frames)
            frames = b''
    stream.stop_stream()
    stream.close()
    audio.terminate()


def live_asr_process(model_name,
                     in_queue,
                     output_queue,
                     beam_width,
                     hotwords,
                     hotword_weight,
                     alpha,
                     beta,
                     beam_prune_logp,
                     token_min_logp,
                     use_auth_token,
                     homophone_extend,
                     kwargs):
    if "tiny" == model_name or "base" == model_name or "small" == model_name or "medium" == model_name and "large" == model_name:
        asr_interface = WhisperInference(model_size=model_name, **kwargs)
    else:
        asr_interface = HFSpeechInference(model_name,
                                          beam_width=beam_width,
                                          hotwords=hotwords,
                                          hotword_weight=hotword_weight,
                                          alpha=alpha,
                                          beta=beta,
                                          beam_prune_logp=beam_prune_logp,
                                          token_min_logp=token_min_logp,
                                          use_auth_token=use_auth_token,
                                          homophone_extend=homophone_extend, **kwargs)

    print("\nlistening to your voice\n")
    while True:
        audio_frames = in_queue.get()
        if audio_frames == "close":
            break

        float64_buffer = np.frombuffer(
            audio_frames, dtype=np.int16).astype(np.float32) / 327678.0
        start = time.perf_counter()
        text = asr_interface.buffer_to_text(float64_buffer).lower()
        inference_time = time.perf_counter() - start
        sample_length = len(float64_buffer) / 16000  # length in sec
        output_queue.put([text, sample_length, inference_time])


def live_get_input_device_id(device_name, microphones):
    for device in microphones:
        if device_name in device[1]:
            return device[0]


def live_list_microphones(pyaudio_instance):
    info = pyaudio_instance.get_host_api_info_by_index(0)
    numdevices = info.get('deviceCount')

    result = []
    for i in range(0, numdevices):
        if (pyaudio_instance.get_device_info_by_host_api_device_index(0, i).get('maxInputChannels')) > 0:
            name = pyaudio_instance.get_device_info_by_host_api_device_index(
                0, i).get('name')
            result += [[i, name]]
    return result


class LiveSpeech:
    exit_event = threading.Event()

    def __init__(self, model_name, device_name="default",
                 vad_mode=2,
                 beam_width=100,
                 hotwords=[],
                 hotword_weight=20,
                 alpha=0.7,
                 beta=1.5,
                 beam_prune_logp=-10.0,
                 token_min_logp=-10.0,
                 use_auth_token=False,
                 homophone_extend=False,
                 **kwargs):
        self.model_name = model_name
        self.device_name = device_name
        self.asr_output_queue = Queue()
        self.asr_input_queue = Queue()
        self.live_asr_process = threading.Thread(target=live_asr_process, args=(
            self.model_name, self.asr_input_queue, self.asr_output_queue,
            beam_width,
            hotwords,
            hotword_weight,
            alpha,
            beta,
            beam_prune_logp,
            token_min_logp,
            use_auth_token,
            homophone_extend,
            kwargs
        ))
        self.live_vad_process = threading.Thread(target=live_vad_process, args=(
            self.device_name, self.asr_input_queue, vad_mode))

    def stop(self):
        """stop the asr process"""
        LiveSpeech.exit_event.set()
        self.asr_input_queue.put("close")
        print("asr stopped")

    def start(self):
        """start the asr process"""
        self.live_asr_process.start()
        time.sleep(3)  # start vad after asr model is loaded
        self.live_vad_process.start()

    def get_last_text(self):
        """returns the text, sample length and inference time in seconds."""
        return self.asr_output_queue.get()



import asrp
import nlp2


# https://github.com/facebookresearch/fairseq/blob/ust/examples/speech_to_speech/docs/textless_s2st_real_data.md
def hifigan_mhubert_en_layer11_code1000():
    nlp2.download_file(
        'https://dl.fbaipublicfiles.com/fairseq/speech_to_speech/vocoder/code_hifigan/mhubert_vp_en_es_fr_it3_400k_layer11_km1000_lj/g_00500000',
        './', 'hifigan_mhubert_en_layer11_code1000_g_00500000')
    nlp2.download_file(
        'https://dl.fbaipublicfiles.com/fairseq/speech_to_speech/vocoder/code_hifigan/mhubert_vp_en_es_fr_it3_400k_layer11_km1000_lj/config.json',
        './', 'hifigan_mhubert_en_layer11_code1000_config.json')
    cs = asrp.Code2Speech(tts_checkpoint='./hifigan_mhubert_en_layer11_code1000_g_00500000',
                          model_cfg=nlp2.read_json('hifigan_mhubert_en_layer11_code1000_config.json'),
                          vocoder='hifigan')
    return cs


def hifigan_mhubert_es_layer11_code1000():
    nlp2.download_file(
        'https://dl.fbaipublicfiles.com/fairseq/speech_to_speech/vocoder/code_hifigan/mhubert_vp_en_es_fr_it3_400k_layer11_km1000_es_css10/g_00500000',
        './', 'hifigan_mhubert_es_layer11_code1000_g_00500000')
    nlp2.download_file(
        'https://dl.fbaipublicfiles.com/fairseq/speech_to_speech/vocoder/code_hifigan/mhubert_vp_en_es_fr_it3_400k_layer11_km1000_es_css10/config.json',
        './', 'hifigan_mhubert_es_layer11_code1000_config.json')
    cs = asrp.Code2Speech(tts_checkpoint='./hifigan_mhubert_es_layer11_code1000_g_00500000',
                          model_cfg=nlp2.read_json('hifigan_mhubert_es_layer11_code1000_config.json'),
                          vocoder='hifigan')
    return cs


def hifigan_mhubert_fr_layer11_code1000():
    nlp2.download_file(
        'https://dl.fbaipublicfiles.com/fairseq/speech_to_speech/vocoder/code_hifigan/mhubert_vp_en_es_fr_it3_400k_layer11_km1000_fr_css10/g_00500000',
        './', 'hifigan_mhubert_fr_layer11_code1000_g_00500000')
    nlp2.download_file(
        'https://dl.fbaipublicfiles.com/fairseq/speech_to_speech/vocoder/code_hifigan/mhubert_vp_en_es_fr_it3_400k_layer11_km1000_fr_css10/config.json',
        './', 'hifigan_mhubert_fr_layer11_code1000_config.json')
    cs = asrp.Code2Speech(tts_checkpoint='./hifigan_mhubert_fr_layer11_code1000_g_00500000',
                          model_cfg=nlp2.read_json('hifigan_mhubert_fr_layer11_code1000_config'),
                          vocoder='hifigan')
    return cs


# https://github.com/facebookresearch/fairseq/tree/ust/examples/speech_matrix
def hifigan_mhubert_de_layer11_code1000():
    nlp2.download_file(
        'https://dl.fbaipublicfiles.com/speech_matrix/vocoder/vocoder_de.pt',
        './')
    nlp2.download_file(
        'https://dl.fbaipublicfiles.com/speech_matrix/vocoder/config_de.json',
        './')
    cs = asrp.Code2Speech(tts_checkpoint='./vocoder_de.pt',
                          model_cfg=nlp2.read_json('config_de.json'),
                          vocoder='hifigan')
    return cs


def hifigan_mhubert_nl_layer11_code1000():
    nlp2.download_file(
        'https://dl.fbaipublicfiles.com/speech_matrix/vocoder/vocoder_nl.pt',
        './')
    nlp2.download_file(
        'https://dl.fbaipublicfiles.com/speech_matrix/vocoder/config_nl.json',
        './')
    cs = asrp.Code2Speech(tts_checkpoint='./vocoder_nl.pt',
                          model_cfg=nlp2.read_json('config_nl.json'),
                          vocoder='hifigan')
    return cs


def hifigan_mhubert_fi_layer11_code1000():
    nlp2.download_file(
        'https://dl.fbaipublicfiles.com/speech_matrix/vocoder/vocoder_fi.pt',
        './')
    nlp2.download_file(
        'https://dl.fbaipublicfiles.com/speech_matrix/vocoder/config_fi.json',
        './')
    cs = asrp.Code2Speech(tts_checkpoint='./vocoder_fi.pt',
                          model_cfg=nlp2.read_json('config_fi.json'),
                          vocoder='hifigan')
    return cs


def hifigan_mhubert_hu_layer11_code1000():
    nlp2.download_file(
        'https://dl.fbaipublicfiles.com/speech_matrix/vocoder/vocoder_hu.pt',
        './')
    nlp2.download_file(
        'https://dl.fbaipublicfiles.com/speech_matrix/vocoder/config_hu.json',
        './')
    cs = asrp.Code2Speech(tts_checkpoint='./vocoder_hu.pt',
                          model_cfg=nlp2.read_json('config_hu.json'),
                          vocoder='hifigan')
    return cs


def hifigan_mhubert_et_layer11_code1000():
    nlp2.download_file(
        'https://dl.fbaipublicfiles.com/speech_matrix/vocoder/vocoder_et.pt',
        './')
    nlp2.download_file(
        'https://dl.fbaipublicfiles.com/speech_matrix/vocoder/config_et.json',
        './')
    cs = asrp.Code2Speech(tts_checkpoint='./vocoder_et.pt',
                          model_cfg=nlp2.read_json('config_et.json'),
                          vocoder='hifigan')
    return cs


def hifigan_mhubert_it_layer11_code800():
    nlp2.download_file(
        'https://dl.fbaipublicfiles.com/speech_matrix/vocoder/vocoder_it.pt',
        './')
    nlp2.download_file(
        'https://dl.fbaipublicfiles.com/speech_matrix/vocoder/config_it.json',
        './')
    cs = asrp.Code2Speech(tts_checkpoint='./vocoder_it.pt',
                          model_cfg=nlp2.read_json('config_it.json'),
                          vocoder='hifigan')
    return cs


def hifigan_mhubert_pt_layer11_code1000():
    nlp2.download_file(
        'https://dl.fbaipublicfiles.com/speech_matrix/vocoder/vocoder_pt.pt',
        './')
    nlp2.download_file(
        'https://dl.fbaipublicfiles.com/speech_matrix/vocoder/config_pt.json',
        './')
    cs = asrp.Code2Speech(tts_checkpoint='./vocoder_pt.pt',
                          model_cfg=nlp2.read_json('config_pt.json'),
                          vocoder='hifigan')
    return cs


def hifigan_mhubert_ro_layer11_code1000():
    nlp2.download_file(
        'https://dl.fbaipublicfiles.com/speech_matrix/vocoder/vocoder_ro.pt',
        './')
    nlp2.download_file(
        'https://dl.fbaipublicfiles.com/speech_matrix/vocoder/config_ro.json',
        './')
    cs = asrp.Code2Speech(tts_checkpoint='./vocoder_ro.pt',
                          model_cfg=nlp2.read_json('config_ro.json'),
                          vocoder='hifigan')
    return cs


def hifigan_mhubert_cs_layer11_code1000():
    nlp2.download_file(
        'https://dl.fbaipublicfiles.com/speech_matrix/vocoder/vocoder_cs.pt',
        './')
    nlp2.download_file(
        'https://dl.fbaipublicfiles.com/speech_matrix/vocoder/config_cs.json',
        './')
    cs = asrp.Code2Speech(tts_checkpoint='./vocoder_cs.pt',
                          model_cfg=nlp2.read_json('config_cs.json'),
                          vocoder='hifigan')
    return cs


def hifigan_mhubert_pl_layer11_code1000():
    nlp2.download_file(
        'https://dl.fbaipublicfiles.com/speech_matrix/vocoder/vocoder_pl.pt',
        './')
    nlp2.download_file(
        'https://dl.fbaipublicfiles.com/speech_matrix/vocoder/config_pl.json',
        './')
    cs = asrp.Code2Speech(tts_checkpoint='./vocoder_pl.pt',
                          model_cfg=nlp2.read_json('config_pl.json'),
                          vocoder='hifigan')
    return cs


def hifigan_mhubert_hr_layer11_code1000():
    nlp2.download_file(
        'https://dl.fbaipublicfiles.com/speech_matrix/vocoder/vocoder_hr.pt',
        './')
    nlp2.download_file(
        'https://dl.fbaipublicfiles.com/speech_matrix/vocoder/config_hr.json',
        './')
    cs = asrp.Code2Speech(tts_checkpoint='./vocoder_hr.pt',
                          model_cfg=nlp2.read_json('config_hr.json'),
                          vocoder='hifigan')
    return cs


def hifigan_mhubert_lt_layer11_code1000():
    nlp2.download_file(
        'https://dl.fbaipublicfiles.com/speech_matrix/vocoder/vocoder_lt.pt',
        './')
    nlp2.download_file(
        'https://dl.fbaipublicfiles.com/speech_matrix/vocoder/config_lt.json',
        './')
    cs = asrp.Code2Speech(tts_checkpoint='./vocoder_lt.pt',
                          model_cfg=nlp2.read_json('config_lt.json'),
                          vocoder='hifigan')
    return cs


def hifigan_mhubert_sk_layer11_code1000():
    nlp2.download_file(
        'https://dl.fbaipublicfiles.com/speech_matrix/vocoder/vocoder_sk.pt',
        './')
    nlp2.download_file(
        'https://dl.fbaipublicfiles.com/speech_matrix/vocoder/config_sk.json',
        './')
    cs = asrp.Code2Speech(tts_checkpoint='./vocoder_sk.pt',
                          model_cfg=nlp2.read_json('config_sk.json'),
                          vocoder='hifigan')
    return cs


def hifigan_mhubert_sl_layer11_code1000():
    nlp2.download_file(
        'https://dl.fbaipublicfiles.com/speech_matrix/vocoder/vocoder_sl.pt',
        './')
    nlp2.download_file(
        'https://dl.fbaipublicfiles.com/speech_matrix/vocoder/config_sl.json',
        './')
    cs = asrp.Code2Speech(tts_checkpoint='./vocoder_sl.pt',
                          model_cfg=nlp2.read_json('config_sl.json'),
                          vocoder='hifigan')
    return cs

# todo: hokkien https://github.com/facebookresearch/fairseq/tree/ust/examples/hokkien



import nlp2

from asrp import HubertCode


def mhubert_layer11_code1000(sampling_rate=16000, chunk_sec=10, worker=8,
                             return_diff=False,
                             batch=None):
    nlp2.download_file(
        'https://huggingface.co/voidful/mhubert-base/resolve/main/mhubert_base_vp_en_es_fr_it3_L11_km1000.bin', './')
    hc = HubertCode("voidful/mhubert-base", './mhubert_base_vp_en_es_fr_it3_L11_km1000.bin', 11,
                    sampling_rate=sampling_rate,
                    chunk_sec=chunk_sec,
                    worker=worker,
                    return_diff=return_diff,
                    batch=batch)
    return hc



from collections import defaultdict
from pathlib import Path

import nlp2


class Seeak:
    def __init__(self, cin_source='jyutping.cin'):
        cin_table_path = nlp2.download_file('https://raw.githubusercontent.com/voidful/asrp/main/data/cin_table.json',
                                            f'{Path(__file__).parent}')
        self.char_to_cin_dict = nlp2.read_json(cin_table_path)
        self.cin_to_char_dict = defaultdict(list)
        self.cin_source = cin_source
        for k, v in self.char_to_cin_dict.items():
            if cin_source in v:
                for cin in v[cin_source]:
                    self.cin_to_char_dict[cin].append(k)
        print("pre calculating homophone character")
        self.homochar = defaultdict(list)
        for char, cins in self.char_to_cin_dict.items():
            for cin in cins.get(cin_source, []):
                self.homochar[char].extend(self.cin_to_char_dict.get(cin, []))

    def get(self, char):
        return self.homochar[char]

    def get_phone(self, char):
        return self.char_to_cin_dict.get(char, {}).get(self.cin_source, [])



import gc
from collections import defaultdict
from functools import partial
from itertools import groupby

import joblib
import numpy as np
import torch
import torchaudio
from torch import nn
from torch.utils.data import DataLoader
from torch.utils.data import Dataset
from tqdm.contrib.concurrent import thread_map
from transformers import Wav2Vec2FeatureExtractor, HubertModel


def collate_fn_pad(batch, device):
    '''
    Padds batch of variable length
    note: it converts things ToTensor manually here since the ToTensor transform
    assume it takes in images rather than arbitrary tensors.
    '''
    ## get sequence lengths
    lengths = torch.tensor([t.shape[0] for t in batch]).to(device)
    ## padd
    batch = [torch.Tensor(t).to(device) for t in batch]
    batch = torch.nn.utils.rnn.pad_sequence(batch, batch_first=True, padding_value=0)
    ## compute mask
    mask = (batch != 0).to(device)
    return batch, lengths, mask


def chunks(l, n):
    """Yield successive n-sized chunks from l."""
    for i in range(0, len(l), n):
        yield l[i:i + n]


class HubertCode(object):
    def __init__(self, hubert_model, km_path, km_layer,
                 sampling_rate=16000,
                 chunk_sec=10,
                 worker=8,
                 return_diff=False,
                 batch=None):
        self.processor = Wav2Vec2FeatureExtractor.from_pretrained(hubert_model)
        self.model = HubertModel.from_pretrained(hubert_model)
        self.model.eval()
        self.sampling_rate = sampling_rate
        self.chunk_length = sampling_rate * chunk_sec
        self.km_model = joblib.load(km_path)
        self.km_layer = km_layer
        self.C_np = self.km_model.cluster_centers_.transpose()
        self.Cnorm_np = (self.C_np ** 2).sum(0, keepdims=True)
        self.worker = worker
        self.return_diff = return_diff
        self.device = 'cuda' if torch.cuda.is_available() else 'cpu'
        self.C = torch.from_numpy(self.C_np)
        self.Cnorm = torch.from_numpy(self.Cnorm_np)
        if torch.cuda.is_available():
            self.C = self.C.cuda()
            self.Cnorm = self.Cnorm.cuda()
            self.model = self.model.cuda()
        self.max_batch = batch if batch else self.get_max_batch()

    def get_max_batch(self):
        print("calculating max batch size...")
        batch = 1
        with torch.no_grad():
            try:
                while True:
                    self.model(torch.rand([batch, self.chunk_length]).cuda())
                    batch += 2
                    gc.collect()
                    torch.cuda.empty_cache()
            except:
                pass
        batch = max(int(batch * 0.95), 1)
        print("maximum batch size will be", batch)
        return batch

    def _process_feature(self, k, top_k=100, feat_norm=False, beamsearch=False, beamsize=5):
        feature = torch.cat(k, dim=0) if isinstance(k, list) else k
        if feat_norm:
            m = nn.BatchNorm1d(feature.shape[-1], affine=False).to(self.device)
            feature = m(feature)
        dist = torch.sqrt(
            feature.pow(2).sum(1, keepdim=True)
            - 2 * torch.matmul(feature, self.C)
            + self.Cnorm
        )
        min_dist = torch.topk(dist.detach().cpu(), top_k, dim=-1, largest=False)
        pred_ind_array = min_dist.indices.cpu().numpy()
        pred_values_array = min_dist.values.cpu().numpy()
        code_output = min_dist.indices.T.cpu().numpy()[0]
        return_dict = {
            'code': list(code_output),
            'merged_code': [k for k, _ in groupby(code_output)]
        }
        if self.return_diff:
            return_dict.update({
                'distance': list(dist.detach().cpu().numpy()),
                'center_diff': list((feature.cpu() - torch.index_select(torch.tensor(self.C_np.transpose()).cpu(), 0,
                                                                        min_dist.indices[:, 0].cpu())).numpy()),
            })
        if beamsearch:
            sequences = [[[], 1.0]]
            self.var_list = []
            for i_row, v_row in zip(pred_ind_array, pred_values_array):
                all_candidates = list()
                for seq in sequences:
                    tokens, score = seq
                    self.var_list.append(np.var(v_row))
                    for k, v in zip(i_row, v_row):
                        norm_len_rate = (len(code_output) / len([k for k, _ in groupby(tokens + [k])]))
                        norm_dist_rate = np.var(v_row) / v
                        candidate = [tokens + [k], score + norm_len_rate * norm_dist_rate]
                        all_candidates.append(candidate)
                ordered = sorted(all_candidates, key=lambda tup: tup[1], reverse=True)
                sequences = ordered[:beamsize]
            code_output = sequences[0][0]
            return_dict['beam_code'] = code_output
            return_dict['beam_merged_code'] = [k for k, _ in groupby(code_output)]
        return return_dict

    def __call__(self, filepaths=None, input_values=None, feat_norm=False, beamsearch=False, top_k=5, beamsize=5):
        with torch.no_grad():
            if filepaths is None:
                filepaths = []

            if input_values is None:
                input_values = []

            if len(filepaths) == 0 and len(input_values) == 0:
                raise ValueError("Both 'filepaths' and 'input_values' are empty. Provide at least one of them.")

            is_single_input = (len(filepaths) == 1 and len(input_values) == 0) or (
                    len(filepaths) == 0 and len(input_values) == 1)

            if isinstance(input_values, torch.Tensor):
                input_values = [input_values]

            if len(filepaths) > 0:
                class SpeechDataset(Dataset):
                    def __init__(self, paths, input_values, processor, sampling_rate=16000):
                        self.paths = paths
                        self.input_values = input_values
                        self.processor = processor
                        self.sampling_rate = sampling_rate

                    def __getitem__(self, index):
                        if index < len(self.paths):
                            speech, sr = torchaudio.load(self.paths[index])
                        else:
                            speech = self.input_values[index - len(self.paths)][None, :]
                            sr = self.sampling_rate

                        speech = speech.mean(0)
                        if sr != self.sampling_rate:
                            resampler = torchaudio.transforms.Resample(orig_freq=sr, new_freq=self.sampling_rate)
                            speech = resampler.forward(speech.squeeze(0))
                        else:
                            speech = speech.squeeze(0)
                        input_values = self.processor(speech, return_tensors="pt",
                                                      sampling_rate=self.sampling_rate).input_values
                        return input_values.squeeze(0)

                    def __len__(self):
                        return len(self.paths) + len(self.input_values)

                def dataloader_collate(batch):
                    return torch.cat(batch, dim=0), [b.shape[0] for b in batch]

                dataset = SpeechDataset(filepaths, input_values, self.processor, self.sampling_rate)
                dataloader = DataLoader(dataset=dataset, batch_size=self.max_batch,
                                        shuffle=False,
                                        num_workers=self.worker,
                                        collate_fn=dataloader_collate)
                process_data_batches = iter(dataloader)
            else:
                def process_input_values():
                    for iv in input_values:
                        yield iv, [iv.shape[0]]

                process_data_batches = process_input_values()

            return_list = []
            for data_batch, size in process_data_batches:
                batch_data = []
                batch_map_audio = []
                for b_id, audio in enumerate(torch.split(data_batch, size)):
                    split_chunks = list(torch.split(audio, self.chunk_length, dim=-1))

                    # Check if the last chunk is smaller than the sampling_rate
                    if split_chunks[-1].shape[-1] < self.sampling_rate:
                        concat_index = -2 if len(split_chunks) >= 2 else 0
                        split_chunks[concat_index] = torch.cat(split_chunks[-2:], dim=-1)
                        split_chunks = split_chunks[:concat_index + 1]

                    # Iterate through chunks and append them to batch_data and batch_map_audio
                    for chunk in split_chunks:
                        batch_data.append(chunk)
                        batch_map_audio.append(b_id)

                code_result = defaultdict(list)
                for bd, bm in zip(chunks(batch_data, self.max_batch), chunks(batch_map_audio, self.max_batch)):
                    batch, lengths, masks = collate_fn_pad(bd, self.device)
                    masks_ratio = lengths / torch.max(lengths)
                    hidden = self.model(batch,
                                        output_hidden_states=True).hidden_states[self.km_layer].detach()
                    mask_len = (hidden.shape[1] * masks_ratio).int()
                    for a, h, ml in zip(bm, hidden, mask_len):
                        code_result[a].append(h[:ml, :])

                for k, v in code_result.items():
                    result = {}
                    for d in thread_map(
                            partial(self._process_feature,
                                    top_k=top_k,
                                    beamsearch=beamsearch,
                                    beamsize=beamsize,
                                    feat_norm=feat_norm), v,
                            leave=False, disable=True):
                        for k2, v2 in d.items():
                            if k2 in result:
                                result[k2].extend(v2)
                            else:
                                result[k2] = v2
                    return_list.append(result)

        if is_single_input:
            return return_list[0]
        else:
            return return_list



import os.path
from typing import Optional, Union, List, Iterable

import numpy as np
import torch

try:
    import whisper
except:
    print("whisper not installed, install it using 'pip install openai-whisper'")
from math import exp
from transformers import AutoModelForCTC, AutoProcessor, WhisperForConditionalGeneration


class HFSpeechInference:
    def __init__(self, model_name, onnx_path='',
                 beam_width=100, hotwords=[], hotword_weight=20,
                 alpha=0.7, beta=1.5,
                 beam_prune_logp=-10.0,
                 token_min_logp=-10.0,
                 use_auth_token=False,
                 homophone_extend=False):
        self.processor = AutoProcessor.from_pretrained(model_name, use_auth_token=use_auth_token)
        self.model = AutoModelForCTC.from_pretrained(model_name, use_auth_token=use_auth_token)
        self.beam_width = beam_width
        self.hotwords = hotwords
        self.hotword_weight = hotword_weight
        self.alpha = alpha
        self.beam_prune_logp = beam_prune_logp
        self.token_min_logp = token_min_logp
        self.beta = beta
        self.is_onnx = False
        if homophone_extend and hasattr(self.processor, 'decoder'):
            self.processor.decoder.enable_homophone_extend = homophone_extend
            print("enable homophone extend")

        if os.path.exists(onnx_path):
            import onnxruntime as rt
            options = rt.SessionOptions()
            options.graph_optimization_level = rt.GraphOptimizationLevel.ORT_ENABLE_ALL
            self.model = rt.InferenceSession(onnx_path, options)
            self.is_onnx = True

    def buffer_to_text(self, audio_buffer):
        if (len(audio_buffer) == 0):
            return ""

        inputs = self.processor(torch.tensor(audio_buffer), sampling_rate=16_000, return_tensors="pt", padding=True)
        with torch.no_grad():
            if self.is_onnx:
                logits = \
                    self.model.run(None, {self.model.get_inputs()[0].name: inputs.input_values.cpu().numpy()})[0]
            else:
                logits = self.model(inputs.input_values).logits.cpu().numpy()

        if hasattr(self.processor, 'decoder'):
            transcription = \
                self.processor.decode(logits[0],
                                      beam_width=self.beam_width,
                                      hotwords=self.hotwords,
                                      hotword_weight=self.hotword_weight,
                                      alpha=self.alpha, beta=self.beta,
                                      token_min_logp=self.token_min_logp,
                                      beam_prune_logp=self.beam_prune_logp,
                                      output_word_offsets=True,
                                      lm_score_boundary=True)
            transcription = transcription.text
        else:
            predicted_ids = np.argmax(logits, axis=-1)
            transcription = self.processor.batch_decode(predicted_ids)[0]

        return transcription.lower()

    def file_to_text(self, filename):
        import librosa
        audio_input, samplerate = librosa.load(filename, sr=16000)
        return self.buffer_to_text(audio_input)


class HFWhisperInference:
    def __init__(self, model_name,
                 beam_width=100, hotwords=[], hotword_weight=20,
                 alpha=0.7, beta=1.5,
                 beam_prune_logp=-10.0,
                 token_min_logp=-10.0,
                 use_auth_token=False):
        self.processor = AutoProcessor.from_pretrained(model_name, use_auth_token=use_auth_token)
        self.model = WhisperForConditionalGeneration.from_pretrained(model_name, use_auth_token=use_auth_token)
        self.beam_width = beam_width
        self.hotwords = hotwords
        self.hotword_weight = hotword_weight
        self.alpha = alpha
        self.beam_prune_logp = beam_prune_logp
        self.token_min_logp = token_min_logp
        self.beta = beta

    def buffer_to_text(self, audio_buffer):
        if (len(audio_buffer) == 0):
            return ""
        inputs = self.processor(torch.tensor(audio_buffer), sampling_rate=16_000, return_tensors="pt")
        with torch.no_grad():
            generated_ids = self.model.generate(inputs=inputs.input_features)
        transcription = self.processor.batch_decode(generated_ids, skip_special_tokens=True)[0]
        return transcription

    def file_to_text(self, filename):
        import librosa
        audio_input, samplerate = librosa.load(filename, sr=16000)
        return self.buffer_to_text(audio_input)


class WhisperInference:
    def __init__(self, model_size='base',
                 task: str = "transcribe",
                 speech_threshold: float = 0.4,
                 temperature: float = 0.0,
                 sample_len: Optional[int] = None,  # maximum number of tokens to sample
                 best_of: Optional[int] = None,  # number of independent samples to collect, when t > 0
                 beam_size: Optional[int] = None,  # number of beams in beam search, when t == 0
                 patience: Optional[float] = None,  # patience in beam search (https://arxiv.org/abs/2204.05424)
                 length_penalty: Optional[float] = None,  # "alpha" in Google NMT, None defaults to length norm
                 prompt: Optional[Union[str, List[int]]] = None,  # text or tokens for the previous context
                 prefix: Optional[Union[str, List[int]]] = None,  # text or tokens to prefix the current context
                 suppress_blank: bool = True,  # this will suppress blank outputs
                 suppress_tokens: Optional[Union[str, Iterable[int]]] = "-1",
                 without_timestamps: bool = False,  # use <|notimestamps|> to sample text tokens only
                 max_initial_timestamp: Optional[float] = 1.0,  # the initial timestamp cannot be later than this
                 language=None):
        self.model = whisper.load_model(model_size)
        self.speech_threshold = speech_threshold
        self.options = whisper.DecodingOptions(fp16=False,
                                               task=task,
                                               language=language,
                                               temperature=temperature,
                                               sample_len=sample_len,
                                               best_of=best_of,
                                               beam_size=beam_size,
                                               patience=patience,
                                               length_penalty=length_penalty,
                                               prompt=prompt,
                                               prefix=prefix,
                                               suppress_blank=suppress_blank,
                                               suppress_tokens=suppress_tokens,
                                               without_timestamps=without_timestamps,
                                               max_initial_timestamp=max_initial_timestamp)

    def buffer_to_text(self, audio_buffer):
        if (len(audio_buffer) == 0):
            return ""

        with torch.no_grad():
            audio = whisper.pad_or_trim(audio_buffer)
            mel = whisper.log_mel_spectrogram(audio).to(self.model.device)
            result = whisper.decode(self.model, mel, self.options)
        if exp(result.avg_logprob) > self.speech_threshold:
            return result.text.lower()
        else:
            return ""

    def file_to_text(self, filename):
        import librosa
        audio_input, samplerate = librosa.load(filename, sr=16000)
        return self.buffer_to_text(audio_input)



import collections
import sys

from asrp import glow
from asrp.hifigan import CodeHiFiGANVocoder

sys.modules['glow'] = glow
import torch
import numpy as np
import torch.nn.functional as F
from torch.autograd import Variable
from scipy.signal import get_window
from asrp.tacotron2 import Tacotron2

device = 'cuda' if torch.cuda.is_available() else 'cpu'


def apply_to_sample(f, sample):
    if hasattr(sample, "__len__") and len(sample) == 0:
        return {}

    def _apply(x):
        if torch.is_tensor(x):
            return f(x)
        elif isinstance(x, collections.OrderedDict):
            # OrderedDict has attributes that needs to be preserved
            od = collections.OrderedDict(
                (key, _apply(value)) for key, value in x.items()
            )
            od.__dict__ = x.__dict__
            return od
        elif isinstance(x, dict):
            return {key: _apply(value) for key, value in x.items()}
        elif isinstance(x, list):
            return [_apply(x) for x in x]
        elif isinstance(x, tuple):
            return tuple(_apply(x) for x in x)
        elif isinstance(x, set):
            return {_apply(x) for x in x}
        else:
            return x

    return _apply(sample)


class Code2Speech(object):
    def __init__(self, tts_checkpoint, model_cfg=None, waveglow_checkpint=None,
                 vocoder='tecotron', max_decoder_steps=2000, end_tok=None, code_begin_pad=0):
        self.device = 'cuda' if torch.cuda.is_available() else 'cpu'
        self.vocoder = vocoder
        if self.vocoder == 'tecotron':
            self.tacotron_model, self.sample_rate, self.hparams = load_tacotron(
                tacotron_model_path=tts_checkpoint,
                max_decoder_steps=max_decoder_steps,
            )
            self.waveglow, self.denoiser = load_waveglow(waveglow_path=waveglow_checkpint)
            self.tacotron_model = self.tacotron_model.to(self.device)
            self.waveglow = self.waveglow.to(self.device)
            self.denoiser = self.denoiser.to(self.device)
        elif self.vocoder == 'hifigan':
            self.sample_rate = 16000
            self.hifigan = load_hifigan(model_path=tts_checkpoint, model_cfg=model_cfg)

        self.end_tok = end_tok
        self.code_begin_pad = code_begin_pad

    def __call__(self, code, strength=0.1, dur_prediction=True):
        with torch.no_grad():
            code = [i + self.code_begin_pad for i in code]
            if self.end_tok is not None and code[-1] != self.end_tok:
                code.append(self.end_tok)
            tts_input = torch.tensor(code, dtype=torch.long)
            if self.vocoder == 'tecotron':
                tts_input.to(device)
                mel, aud, aud_dn, has_eos = synthesize_audio(
                    self.tacotron_model,
                    self.waveglow,
                    self.denoiser,
                    tts_input.unsqueeze(0),
                    strength=strength
                )
                audio_seq = aud_dn[0].cpu().float().numpy()
            elif self.vocoder == 'hifigan':
                x = {
                    "code": tts_input.view(1, -1)
                }
                # if self.hifigan.multispkr:
                #     spk = (
                #         random.randint(0, vocnum_speakers - 1)
                #         if args.speaker_id == -1
                #         else args.speaker_id
                #     )
                #     suffix = f"_spk{spk}"
                #     x["spkr"] = torch.LongTensor([spk]).view(1, 1)
                audio_seq = self.hifigan(x, dur_prediction=dur_prediction)

            return audio_seq


def window_sumsquare(window, n_frames, hop_length=200, win_length=800,
                     n_fft=800, dtype=np.float32, norm=None):
    """
    # from librosa 0.6
    Compute the sum-square envelope of a window function at a given hop length.
    This is used to estimate modulation effects induced by windowing
    observations in short-time fourier transforms.
    Parameters
    ----------
    window : string, tuple, number, callable, or list-like
        Window specification, as in `get_window`
    n_frames : int > 0
        The number of analysis framesa
    hop_length : int > 0
        The number of samples to advance between frames
    win_length : [optional]
        The length of the window function.  By default, this matches `n_fft`.
    n_fft : int > 0
        The length of each analysis frame.
    dtype : np.dtype
        The data type of the output
    Returns
    -------
    wss : np.ndarray, shape=`(n_fft + hop_length * (n_frames - 1))`
        The sum-squared envelope of the window function
    """
    import librosa
    if win_length is None:
        win_length = n_fft

    n = n_fft + hop_length * (n_frames - 1)
    x = np.zeros(n, dtype=dtype)

    # Compute the squared window at the desired length
    win_sq = get_window(window, win_length, fftbins=True)
    win_sq = librosa.util.normalize(win_sq, norm=norm) ** 2
    win_sq = librosa.util.pad_center(win_sq, n_fft)

    # Fill the envelope
    for i in range(n_frames):
        sample = i * hop_length
        x[sample:min(n, sample + n_fft)] += win_sq[:max(0, min(n_fft, n - sample))]
    return x


class STFT(torch.nn.Module):
    """adapted from Prem Seetharaman's https://github.com/pseeth/pytorch-stft"""

    def __init__(self, filter_length=800, hop_length=200, win_length=800,
                 window='hann'):
        super(STFT, self).__init__()
        self.filter_length = filter_length
        self.hop_length = hop_length
        self.win_length = win_length
        self.window = window
        self.forward_transform = None
        scale = self.filter_length / self.hop_length
        fourier_basis = np.fft.fft(np.eye(self.filter_length))

        cutoff = int((self.filter_length / 2 + 1))
        fourier_basis = np.vstack([np.real(fourier_basis[:cutoff, :]),
                                   np.imag(fourier_basis[:cutoff, :])])

        forward_basis = torch.FloatTensor(fourier_basis[:, None, :])
        inverse_basis = torch.FloatTensor(
            np.linalg.pinv(scale * fourier_basis).T[:, None, :])

        if window is not None:
            import librosa
            assert (filter_length >= win_length)
            # get window and zero center pad it to filter_length
            fft_window = get_window(window, win_length, fftbins=True)
            fft_window = librosa.util.pad_center(fft_window, filter_length)
            fft_window = torch.from_numpy(fft_window).float()

            # window the bases
            forward_basis *= fft_window
            inverse_basis *= fft_window

        self.register_buffer('forward_basis', forward_basis.float())
        self.register_buffer('inverse_basis', inverse_basis.float())

    def transform(self, input_data):
        num_batches = input_data.size(0)
        num_samples = input_data.size(1)

        self.num_samples = num_samples

        # similar to librosa, reflect-pad the input
        input_data = input_data.view(num_batches, 1, num_samples)
        input_data = F.pad(
            input_data.unsqueeze(1),
            (int(self.filter_length / 2), int(self.filter_length / 2), 0, 0),
            mode='reflect')
        input_data = input_data.squeeze(1)

        forward_transform = F.conv1d(
            input_data,
            Variable(self.forward_basis, requires_grad=False),
            stride=self.hop_length,
            padding=0)

        cutoff = int((self.filter_length / 2) + 1)
        real_part = forward_transform[:, :cutoff, :]
        imag_part = forward_transform[:, cutoff:, :]

        magnitude = torch.sqrt(real_part ** 2 + imag_part ** 2)
        phase = torch.autograd.Variable(
            torch.atan2(imag_part.data, real_part.data))

        return magnitude, phase

    def inverse(self, magnitude, phase):
        recombine_magnitude_phase = torch.cat(
            [magnitude * torch.cos(phase), magnitude * torch.sin(phase)], dim=1)

        inverse_transform = F.conv_transpose1d(
            recombine_magnitude_phase,
            Variable(self.inverse_basis, requires_grad=False),
            stride=self.hop_length,
            padding=0)

        if self.window is not None:
            import librosa
            window_sum = window_sumsquare(
                self.window, magnitude.size(-1), hop_length=self.hop_length,
                win_length=self.win_length, n_fft=self.filter_length,
                dtype=np.float32)
            # remove modulation effects
            approx_nonzero_indices = torch.from_numpy(
                np.where(window_sum > librosa.util.tiny(window_sum))[0])
            window_sum = torch.autograd.Variable(
                torch.from_numpy(window_sum), requires_grad=False)
            window_sum = window_sum.cuda() if magnitude.is_cuda else window_sum
            inverse_transform[:, :, approx_nonzero_indices] /= window_sum[approx_nonzero_indices]

            # scale by hop ratio
            inverse_transform *= float(self.filter_length) / self.hop_length

        inverse_transform = inverse_transform[:, :, int(self.filter_length / 2):]
        inverse_transform = inverse_transform[:, :, :-int(self.filter_length / 2):]

        return inverse_transform

    def forward(self, input_data):
        self.magnitude, self.phase = self.transform(input_data)
        reconstruction = self.inverse(self.magnitude, self.phase)
        return reconstruction


class Denoiser(torch.nn.Module):
    """ Removes model bias from audio produced with waveglow """

    def __init__(self, waveglow, filter_length=1024, n_overlap=4,
                 win_length=1024, mode='zeros'):
        super(Denoiser, self).__init__()
        self.device = 'cuda' if torch.cuda.is_available() else 'cpu'
        self.stft = STFT(filter_length=filter_length,
                         hop_length=int(filter_length / n_overlap),
                         win_length=win_length).to(self.device)
        if mode == 'zeros':
            mel_input = torch.zeros(
                (1, 80, 88),
                dtype=waveglow.upsample.weight.dtype,
                device=waveglow.upsample.weight.device)
        elif mode == 'normal':
            mel_input = torch.randn(
                (1, 80, 88),
                dtype=waveglow.upsample.weight.dtype,
                device=waveglow.upsample.weight.device)
        else:
            raise Exception("Mode {} if not supported".format(mode))

        with torch.no_grad():
            bias_audio = waveglow.infer(mel_input, sigma=0.0).float()
            bias_spec, _ = self.stft.transform(bias_audio)

        self.register_buffer('bias_spec', bias_spec[:, :, 0][:, :, None])

    def forward(self, audio, strength=0.1):
        audio_spec, audio_angles = self.stft.transform(audio.to(self.device).float())
        audio_spec_denoised = audio_spec - self.bias_spec * strength
        audio_spec_denoised = torch.clamp(audio_spec_denoised, 0.0)
        audio_denoised = self.stft.inverse(audio_spec_denoised, audio_angles)
        return audio_denoised


def load_quantized_audio_from_file(file_path):
    base_fname_batch, quantized_units_batch = [], []
    with open(file_path) as f:
        for line in f:
            base_fname, quantized_units_str = line.rstrip().split("|")
            quantized_units = [int(q) for q in quantized_units_str.split(" ")]
            base_fname_batch.append(base_fname)
            quantized_units_batch.append(quantized_units)
    return base_fname_batch, quantized_units_batch


def synthesize_audio(model, waveglow, denoiser, inp, lab=None, strength=0.0):
    assert inp.size(0) == 1
    inp = inp.to(device)
    if lab is not None:
        lab = torch.LongTensor(1).to(device).fill_(lab)

    with torch.no_grad():
        _, mel, _, ali, has_eos = model.inference(inp, lab, ret_has_eos=True)
        aud = waveglow.infer(mel, sigma=0.666)
        aud_dn = denoiser(aud, strength=strength).squeeze(1)
    return mel, aud, aud_dn, has_eos


def load_tacotron(tacotron_model_path, max_decoder_steps):
    ckpt_dict = torch.load(tacotron_model_path, map_location=torch.device(device))
    hparams = ckpt_dict["hparams"]
    hparams.max_decoder_steps = max_decoder_steps
    sr = hparams.sampling_rate
    model = Tacotron2(hparams)
    model.load_state_dict(ckpt_dict["model_dict"])
    model = model.eval().half()
    return model, sr, hparams


def load_waveglow(waveglow_path):
    if waveglow_path is None:
        waveglow = torch.hub.load('NVIDIA/DeepLearningExamples:torchhub', 'nvidia_waveglow', model_math='fp32')
    else:
        waveglow = torch.load(waveglow_path, map_location=torch.device(device))["model"]

    waveglow = waveglow.remove_weightnorm(waveglow)
    waveglow = waveglow.eval().half().to(device)
    for k in waveglow.convinv:
        k.float()
    denoiser = Denoiser(waveglow)
    return waveglow, denoiser


def load_hifigan(model_path, model_cfg=None, speaker_id=0, num_speakers=200):
    vocoder = CodeHiFiGANVocoder(model_path, model_cfg)
    multispkr = vocoder.model.multispkr
    if multispkr:
        assert (
                speaker_id < num_speakers
        ), f"invalid --speaker-id ({speaker_id}) with total #speakers = {num_speakers}"
    vocoder.multispkr = multispkr
    vocoder.speaker_id = speaker_id
    vocoder.num_speakers = num_speakers
    return vocoder



import os
from pathlib import Path

import nlp2
import torch
import torchaudio


def extract_d_vector(wav_path=None, wav_tensor=None, sampling_rate=16000, inference=True):
    device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
    with torch.inference_mode(inference):
        try:
            save_folder = Path(__file__).parent
        except:
            save_folder = './'
        wav2mel_path = os.path.join(save_folder, 'wav2mel.pt')
        dvector_path = os.path.join(save_folder, 'dvector-step250000.pt')
        if not os.path.exists(wav2mel_path):
            nlp2.download_file(
                'https://github.com/yistLin/dvector/releases/download/v1.1.1/wav2mel.pt',
                f'{Path(__file__).parent}')
        if not os.path.exists(dvector_path):
            nlp2.download_file(
                'https://github.com/yistLin/dvector/releases/download/v1.1.1/dvector-step250000.pt',
                f'{Path(__file__).parent}')
        wav2mel = torch.jit.load(wav2mel_path)
        dvector = torch.jit.load(dvector_path).eval().to(device)
        if not torch.is_tensor(wav_tensor):
            wav_tensor, sample_rate = torchaudio.load(wav_path)
            mel_tensor = wav2mel(wav_tensor, sample_rate)  # shape: (frames, mel_dim)
        else:
            mel_tensor = wav2mel(wav_tensor, sampling_rate)
        emb_tensor = dvector.embed_utterance(mel_tensor.to(device))  # shape: (emb_dim)
    return emb_tensor.tolist()


def extract_x_vector(wav_path=None, wav_tensor=None, sampling_rate=16000, inference=True):
    device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
    with torch.inference_mode(inference):
        from speechbrain.pretrained import EncoderClassifier
        classifier = EncoderClassifier.from_hparams(source="speechbrain/spkrec-xvect-voxceleb",
                                                    savedir="pretrained_models/spkrec-xvect-voxceleb",
                                                    run_opts={"device": device})
        if not torch.is_tensor(wav_tensor):
            wav_tensor, sampling_rate = torchaudio.load(wav_path)
        embeddings = classifier.encode_batch(wav_tensor.to(device)).squeeze()
        return embeddings.tolist()



# Copyright (c) Facebook, Inc. and its affiliates.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.
# author: adefossez

import math

import torch as th
from torch.nn import functional as F


def sinc(t):
    """sinc.

    :param t: the input tensor
    """
    return th.where(t == 0, th.tensor(1., device=t.device, dtype=t.dtype),
                    th.sin(t) / t)


def kernel_upsample2(zeros=56):
    """kernel_upsample2.

    """
    win = th.hann_window(4 * zeros + 1, periodic=False)
    winodd = win[1::2]
    t = th.linspace(-zeros + 0.5, zeros - 0.5, 2 * zeros)
    t *= math.pi
    kernel = (sinc(t) * winodd).view(1, 1, -1)
    return kernel


def upsample2(x, zeros=56):
    """
    Upsampling the input by 2 using sinc interpolation.
    Smith, Julius, and Phil Gossett. "A flexible sampling-rate conversion method."
    ICASSP'84. IEEE International Conference on Acoustics, Speech, and Signal Processing.
    Vol. 9. IEEE, 1984.
    """
    *other, time = x.shape
    kernel = kernel_upsample2(zeros).to(x)
    out = F.conv1d(x.view(-1, 1, time), kernel, padding=zeros)[..., 1:].view(
        *other, time
    )
    y = th.stack([x, out], dim=-1)
    return y.view(*other, -1)


def kernel_downsample2(zeros=56):
    """kernel_downsample2.

    """
    win = th.hann_window(4 * zeros + 1, periodic=False)
    winodd = win[1::2]
    t = th.linspace(-zeros + 0.5, zeros - 0.5, 2 * zeros)
    t.mul_(math.pi)
    kernel = (sinc(t) * winodd).view(1, 1, -1)
    return kernel


def downsample2(x, zeros=56):
    """
    Downsampling the input by 2 using sinc interpolation.
    Smith, Julius, and Phil Gossett. "A flexible sampling-rate conversion method."
    ICASSP'84. IEEE International Conference on Acoustics, Speech, and Signal Processing.
    Vol. 9. IEEE, 1984.
    """
    if x.shape[-1] % 2 != 0:
        x = F.pad(x, (0, 1))
    xeven = x[..., ::2]
    xodd = x[..., 1::2]
    *other, time = xodd.shape
    kernel = kernel_downsample2(zeros).to(x)
    out = xeven + F.conv1d(
        xodd.view(-1, 1, time), kernel, padding=zeros
    )[..., :-1].view(*other, time)
    return out.view(*other, -1).mul(0.5)



import torch
import torchaudio


class s3prlModel:
    def __init__(self, cpkt, dict_path='dict.txt'):
        from s3prl.downstream.runner import Runner
        model_dict = torch.load(cpkt, map_location='cpu')
        self.args = model_dict['Args']
        self.config = model_dict['Config']
        # patch the config
        self.args.init_ckpt = cpkt
        self.args.device = "cuda" if torch.cuda.is_available() else "cpu"
        self.config['downstream_expert']['datarc']["dict_path"] = dict_path
        runner = Runner(self.args, self.config)
        self.upstream = runner._get_upstream()
        self.featurizer = runner._get_featurizer()
        self.downstream = runner._get_downstream()

    def __call__(self, wav_path):
        wav, sr = torchaudio.load(wav_path)
        wav = wav.mean(0).unsqueeze(0)
        with torch.no_grad():
            features = self.upstream.model(wav)
            features = self.featurizer.model(wav, features)
            result = self.downstream.model.inference(features, filenames=[])[0]
            return result


s = s3prlModel('./backup/dev-clean-best.ckpt', './backup/char.dict')
print(s('./backup/sample1.flac'))
#["GOING ALONG SLUSHY COUNTRI ROADS AND SPEAKING TO DAMP ORDIANC IS IN DROFTY'S CHOOL ROMS DAY OF TER DAY FOR FOULT NIGHT HE'LL HAVE TO PUT IN AN APPEARANCE OF SUMONG PLACE OF WUSH IT POND SOME DAY MONING AND HE CAN COME TO USSI MEDIATELY OFTOWAODS"]


import re
import string
import unicodedata

langs = ['ab', 'ar', 'as', 'br', 'ca', 'cnh', 'cs', 'cv', 'cy', 'de', 'dv', 'el', 'en', 'eo', 'es', 'et', 'eu', 'fa',
         'fi', 'fr', 'fy-NL', 'ga-IE', 'hi', 'hsb', 'hu', 'ia', 'id', 'it', 'ja', 'ka', 'kab', 'ky', 'lg', 'lt', 'lv',
         'mn', 'mt', 'nl', 'or', 'pa-IN', 'pl', 'pt', 'rm-sursilv', 'rm-vallader', 'ro', 'ru', 'rw', 'sah', 'sl',
         'sv-SE', 'ta', 'th', 'tr', 'tt', 'uk', 'vi', 'vot', 'zh-CN', 'zh-HK', 'zh-TW']


def fun_ar(batch):
    chars_to_ignore_regex = '[\\\\\\\\\\\\\\\\؛\\\\\\\\\\\\\\\\—\\\\\\\\\\\\\\\\_get\\\\\\\\\\\\\\\\«\\\\\\\\\\\\\\\\»\\\\\\\\\\\\\\\\ـ\\\\\\\\\\\\\\\\ـ\\\\\\\\\\\\\\\\,\\\\\\\\\\\\\\\\?\\\\\\\\\\\\\\\\.\\\\\\\\\\\\\\\\!\\\\\\\\\\\\\\\\-\\\\\\\\\\\\\\\\;\\\\\\\\\\\\\\\\:\\\\\\\\\\\\\\\\"\\\\\\\\\\\\\\\\“\\\\\\\\\\\\\\\\%\\\\\\\\\\\\\\\\‘\\\\\\\\\\\\\\\\”\\\\\\\\\\\\\\\\�\\\\\\\\\\\\\\\\#\\\\\\\\\\\\\\\\،\\\\\\\\\\\\\\\\☭,\\\\\\\\\\\\\\\\؟]'
    batch["sentence"] = re.sub(chars_to_ignore_regex, '', batch["sentence"]).lower()
    return batch


def fun_as(batch):
    chars_to_ignore_regex = '[\\,\\?\\.\\!\\-\\;\\:\\"\\“\\%\\”\\়\\।]'
    batch["sentence"] = re.sub('’ ', ' ', batch["sentence"])
    batch["sentence"] = re.sub(' ‘', ' ', batch["sentence"])
    batch["sentence"] = re.sub('’|‘', '\'', batch["sentence"])
    batch["sentence"] = re.sub(chars_to_ignore_regex, '', batch["sentence"]).lower()
    return batch


def fun_br(batch):
    chars_to_ignore_regex = '[\\,\,\?\.\!\;\:\"\“\%\”\�\(\)\/\«\»\½\…]'
    batch["sentence"] = re.sub(chars_to_ignore_regex, '', batch["sentence"]).lower()
    batch["sentence"] = re.sub(chars_to_ignore_regex, '', batch["sentence"]).lower() + " "
    batch["sentence"] = re.sub("ʼ", "'", batch["sentence"])
    batch["sentence"] = re.sub("’", "'", batch["sentence"])
    batch["sentence"] = re.sub('‘', "'", batch["sentence"])
    return batch


def fun_ca(batch):
    chars_to_ignore_regex = '[\,\?\.\!\;\:\"\“]'
    batch["sentence"] = re.sub(chars_to_ignore_regex, '', batch["sentence"]).lower()
    return batch


def fun_cnh(batch):
    chars_to_ignore_regex = '[\,\?\.\!\-\;\:\"\“\%\‘\”\�\/]'
    batch["sentence"] = re.sub(chars_to_ignore_regex, '', batch["sentence"]).lower()
    return batch


def fun_cs(batch):
    chars_to_ignore = [",", "?", ".", "!", "-", ";", ":", '""', "%", "'", '"', "�", '«', '»', '—', '…', '(', ')', '*',
                       '”', '“']
    chars_to_ignore_regex = f'[{"".join(chars_to_ignore)}]'
    batch["sentence"] = re.sub(chars_to_ignore_regex, '', batch["sentence"]).lower().strip()
    batch["sentence"] = re.sub(re.compile('[äá]'), 'a', batch['sentence'])
    batch["sentence"] = re.sub(re.compile('[öó]'), 'o', batch['sentence'])
    batch["sentence"] = re.sub(re.compile('[èé]'), 'e', batch['sentence'])
    batch["sentence"] = re.sub(re.compile("[ïí]"), 'i', batch['sentence'])
    batch["sentence"] = re.sub(re.compile("[üů]"), 'u', batch['sentence'])
    batch['sentence'] = re.sub('  ', ' ', batch['sentence'])
    return batch


def fun_cv(batch):
    sent = batch["sentence"].lower()
    # replace non-alpha characters with space
    sent = "".join(ch if ch.isalpha() else " " for ch in sent)
    # remove repeated spaces
    sent = " ".join(sent.split())
    batch["sentence"] = sent
    return batch


def fun_cy(batch):
    chars_to_ignore_regex = '[\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\,\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\?\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\.\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\!\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\-\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\u2013\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\u2014\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\;\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\:\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\"\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\%\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\]'
    batch["sentence"] = re.sub(chars_to_ignore_regex, '', batch["sentence"]).lower()
    return batch


def fun_de(batch):
    chars_to_ignore_regex = '[\\,\\?\\.\\!\\-\\;\\:\\"\\“]'
    batch["sentence"] = re.sub(chars_to_ignore_regex, '', batch["sentence"]).lower()
    return batch


def fun_dv(batch):
    chars_to_ignore_regex = '[\,\?\.\!\-\;\:\"\“\%\‘\”\�\،\.\؟\!\'\"\–\’]'
    batch["sentence"] = re.sub(chars_to_ignore_regex, '', batch["sentence"]).lower()
    return batch


def fun_el(batch):
    chars_to_ignore_regex = '[\\\\\\\\,\\\\\\\\?\\\\\\\\.\\\\\\\\!\\\\\\\\-\\\\\\\\;\\\\\\\\:\\\\\\\\"\\\\\\\\“\\\\\\\\%\\\\\\\\‘\\\\\\\\”\\\\\\\\�]'
    batch["sentence"] = re.sub(chars_to_ignore_regex, '', batch["sentence"]).lower()
    return batch


def fun_en(batch):
    sent = batch["sentence"].lower()
    # normalize apostrophes
    sent = sent.replace("’", "'")
    # replace non-alpha characters with space
    sent = "".join(ch if ch.isalpha() or ch == "'" else " " for ch in sent)
    # remove repeated spaces
    sent = " ".join(sent.split())
    batch["sentence"] = sent
    return batch


def fun_eo(batch):
    chars_to_ignore_regex = """[\\\\\\\\,\\\\\\\\?\\\\\\\\.\\\\\\\\!\\\\\\\\-\\\\\\\\;\\\\\\\\:\\\\\\\\"\\\\\\\\“\\\\\\\\%\\\\\\\\‘\\\\\\\\”\\\\\\\\�\\\\\\\\„\\\\\\\\«\\\\\\\\(\\\\\\\\»\\\\\\\\)\\\\\\\\’\\\\\\\\']"""
    batch["sentence"] = re.sub(chars_to_ignore_regex, '', batch["sentence"]).lower().replace('—', ' ').replace('–', ' ')
    return batch


def fun_es(batch):
    # remove_special_characters
    chars_to_ignore_regex = '[\,\¿\?\.\¡\!\-\;\:\"\“\%\‘\”\￼\…\’\ː\'\‹\›\`\´\®\—\→]'
    chars_to_ignore_pattern = re.compile(chars_to_ignore_regex)
    batch["sentence"] = chars_to_ignore_pattern.sub('', batch["sentence"]).lower() + " "
    # replace_diacritics
    sentence = batch["sentence"]
    sentence = re.sub('ì', 'í', sentence)
    sentence = re.sub('ù', 'ú', sentence)
    sentence = re.sub('ò', 'ó', sentence)
    sentence = re.sub('à', 'á', sentence)
    batch["sentence"] = sentence
    # replace_additional
    sentence = batch["sentence"]
    sentence = re.sub('ã', 'a', sentence)  # Portuguese, as in São Paulo
    sentence = re.sub('ō', 'o', sentence)  # Japanese
    sentence = re.sub('ê', 'e', sentence)  # Português
    batch["sentence"] = sentence
    return batch


def fun_et(batch):
    sent = batch["sentence"].lower()
    # normalize apostrophes
    sent = sent.replace("’", "'")
    # replace non-alpha characters with space
    sent = "".join(ch if ch.isalpha() or ch == "'" else " " for ch in sent)
    # remove repeated spaces
    sent = " ".join(sent.split())
    batch["sentence"] = sent
    return batch


def fun_eu(batch):
    chars_to_ignore_regex = '[\,\¿\?\.\¡\!\-\;\:\"\“\%\‘\”\￼\…\’\ː\'\‹\›\`\´\®\—\→]'
    chars_to_ignore_pattern = re.compile(chars_to_ignore_regex)
    batch["sentence"] = chars_to_ignore_pattern.sub('', batch["sentence"]).lower() + " "
    return batch


def fun_fa(batch):
    import hazm
    _normalizer = hazm.Normalizer()
    chars_to_ignore = [
        ",", "?", ".", "!", "-", ";", ":", '""', "%", "'", '"', "�",
        "#", "!", "؟", "?", "«", "»", "،", "(", ")", "؛", "'ٔ", "٬", 'ٔ', ",", "?",
        ".", "!", "-", ";", ":", '"', "“", "%", "‘", "”", "�", "–", "…", "_", "”", '“', '„',
        'ā', 'š',
        # "ء",
    ]
    chars_to_ignore = chars_to_ignore + list(string.ascii_lowercase + string.digits)
    chars_to_mapping = {
        'ك': 'ک', 'دِ': 'د', 'بِ': 'ب', 'زِ': 'ز', 'ذِ': 'ذ', 'شِ': 'ش', 'سِ': 'س', 'ى': 'ی',
        'ي': 'ی', 'أ': 'ا', 'ؤ': 'و', "ے": "ی", "ۀ": "ه", "ﭘ": "پ", "ﮐ": "ک", "ﯽ": "ی",
        "ﺎ": "ا", "ﺑ": "ب", "ﺘ": "ت", "ﺧ": "خ", "ﺩ": "د", "ﺱ": "س", "ﻀ": "ض", "ﻌ": "ع",
        "ﻟ": "ل", "ﻡ": "م", "ﻢ": "م", "ﻪ": "ه", "ﻮ": "و", 'ﺍ': "ا", 'ة': "ه",
        'ﯾ': "ی", 'ﯿ': "ی", 'ﺒ': "ب", 'ﺖ': "ت", 'ﺪ': "د", 'ﺮ': "ر", 'ﺴ': "س", 'ﺷ': "ش",
        'ﺸ': "ش", 'ﻋ': "ع", 'ﻤ': "م", 'ﻥ': "ن", 'ﻧ': "ن", 'ﻭ': "و", 'ﺭ': "ر", "ﮔ": "گ",

        # "ها": "  ها", "ئ": "ی",

        "a": " ای ", "b": " بی ", "c": " سی ", "d": " دی ", "e": " ایی ", "f": " اف ",
        "g": " جی ", "h": " اچ ", "i": " آی ", "j": " جی ", "k": " کی ", "l": " ال ",
        "m": " ام ", "n": " ان ", "o": " او ", "p": " پی ", "q": " کیو ", "r": " آر ",
        "s": " اس ", "t": " تی ", "u": " یو ", "v": " وی ", "w": " دبلیو ", "x": " اکس ",
        "y": " وای ", "z": " زد ",
        "\u200c": " ", "\u200d": " ", "\u200e": " ", "\u200f": " ", "\ufeff": " ",
    }
    chars_to_ignore_regex = f"""[{"".join(chars_to_ignore)}]"""
    text = batch["sentence"].lower().strip()

    text = _normalizer.normalize(text)
    # multiple_replace
    pattern = "|".join(map(re.escape, chars_to_mapping.keys()))
    text = re.sub(pattern, lambda m: chars_to_mapping[m.group()], str(text))

    # remove_special_characters
    text = re.sub(chars_to_ignore_regex, '', text).lower() + " "

    text = re.sub(" +", " ", text)
    text = text.strip() + " "

    batch["sentence"] = text
    return batch


def fun_fi(batch):
    chars_to_ignore_regex = '[\,\?\.\!\-\;\:\"\“\%\‘\”\�\'\...\…\–\é]'
    batch["sentence"] = re.sub(chars_to_ignore_regex, '', batch["sentence"]).lower()
    return batch


def fun_fr(batch):
    chars_to_ignore_regex = '[\,\?\.\!\-\;\:\"\“\%\‘\”\�]'
    batch["sentence"] = re.sub(chars_to_ignore_regex, '', batch["sentence"]).lower()
    return batch


def fun_fy_NL(batch):
    chars_to_ignore_regex = '[\,\?\.\!\-\;\:\"\'\“\%\‘\”]'
    batch["sentence"] = re.sub(chars_to_ignore_regex, '', batch["sentence"]).lower()
    return batch


def fun_ga_IE(batch):
    chars_to_ignore_regex = '[\,\?\.\!\-\;\:\"\“\%\‘\”\�\’\–\(\)]'
    batch["sentence"] = re.sub(chars_to_ignore_regex, '', batch["sentence"]).lower()
    return batch


def fun_hi(batch):
    chars_to_ignore_regex = '[\,\?\.\!\-\;\:\"\“\%\‘\”\�\।]'
    batch["sentence"] = re.sub(chars_to_ignore_regex, '', batch["sentence"])
    return batch


def fun_hsb(batch):
    chars_to_ignore_regex = '[\\\\\\\\\\\\\\\\,\\\\\\\\\\\\\\\\?\\\\\\\\\\\\\\\\.\\\\\\\\\\\\\\\\!\\\\\\\\\\\\\\\\-\\\\\\\\\\\\\\\\;\\\\\\\\\\\\\\\\:\\\\\\\\\\\\\\\\"\\\\\\\\\\\\\\\\“\\\\\\\\\\\\\\\\%\\\\\\\\\\\\\\\\‘\\\\\\\\\\\\\\\\”\\\\\\\\\\\\\\\\�\\\\\\\\\\\\\\\\–\\\\\\\\\\\\\\\\—\\\\\\\\\\\\\\\\¬\\\\\\\\\\\\\\\\⅛]'
    batch["sentence"] = re.sub(chars_to_ignore_regex, '', batch["sentence"]).lower()
    return batch


def fun_hu(batch):
    CHARS_TO_IGNORE = [",", "?", "¿", ".", "!", "¡", ";", ":", '""', "%", '"', "�", "ʿ", "·", "჻", "~", "՞",
                       "؟", "،", "।", "॥", "«", "»", "„", "“", "”", "「", "」", "‘", "’", "《", "》", "(", ")", "[", "]",
                       "=", "`", "_", "+", "<", ">", "…", "–", "°", "´", "ʾ", "‹", "›", "©", "®", "—", "→", "。"]
    chars_to_ignore_regex = f"[{re.escape(''.join(CHARS_TO_IGNORE))}]"
    batch["sentence"] = re.sub(chars_to_ignore_regex, "", batch["sentence"]).upper()
    return batch


def fun_ia(batch):
    chars_to_ignore_regex = '[\.\,\!\?\-\"\:\;\'\“\”]'
    batch["sentence"] = re.sub(chars_to_ignore_regex, '', batch["sentence"]).lower()
    return batch


def fun_id(batch):
    chars_to_ignore_regex = '[\,\?\.\!\-\;\:\"\“\%\‘\'\”\�]'
    batch["sentence"] = re.sub(chars_to_ignore_regex, '', batch["sentence"]).lower()
    return batch


def fun_it(batch):
    chars_to_remove = [",", "?", ".", "!", "-", ";", ":", '""', "%", '"', "�", 'ʿ', '“', '”', '(', '=', '`', '_', '+',
                       '«', '<', '>', '~', '…', '«', '»', '–', '\[', '\]', '°', '̇', '´', 'ʾ', '„', '̇', '̇', '̇',
                       '¡']  # All extra characters

    chars_to_remove_regex = f'[{"".join(chars_to_remove)}]'
    batch["sentence"] = re.sub(chars_to_remove_regex, '', batch["sentence"]).lower().replace('‘', "'").replace('ʻ',
                                                                                                               "'").replace(
        'ʼ', "'").replace('’', "'").replace('ʹ', "''").replace('̇', '')

    allowed_characters = [
        " ",
        "'",
        'a',
        'b',
        'c',
        'd',
        'e',
        'f',
        'g',
        'h',
        'i',
        'j',
        'k',
        'l',
        'm',
        'n',
        'o',
        'p',
        'q',
        'r',
        's',
        't',
        'u',
        'v',
        'w',
        'x',
        'y',
        'z',
        'à',
        'á',
        'è',
        'é',
        'ì',
        'í',
        'ò',
        'ó',
        'ù',
        'ú',
    ]

    def remove_accents(input_str):
        if input_str in allowed_characters:
            return input_str
        if input_str == 'ø':
            return 'o'
        elif input_str == 'ß' or input_str == 'ß':
            return 'b'
        elif input_str == 'ё':
            return 'e'
        elif input_str == 'đ':
            return 'd'
        nfkd_form = unicodedata.normalize('NFKD', input_str)
        only_ascii = nfkd_form.encode('ASCII', 'ignore').decode()

        if only_ascii is None or only_ascii == '':
            return input_str
        else:
            return only_ascii

    def fix_accents(sentence):
        new_sentence = ''
        for char in sentence:
            new_sentence += remove_accents(char)
        return new_sentence

    batch["sentence"] = fix_accents(batch["sentence"])
    return batch


def fun_ja(batch):
    CHARS_TO_IGNORE = [",", "?", "¿", ".", "!", "¡", ";", ":", '""', "%", '"', "�", "ʿ", "·", "჻", "~", "՞",
                       "؟", "،", "।", "॥", "«", "»", "„", "“", "”", "「", "」", "‘", "’", "《", "》", "(", ")", "[", "]",
                       "=", "`", "_", "+", "<", ">", "…", "–", "°", "´", "ʾ", "‹", "›", "©", "®", "—", "→", "。"]
    chars_to_ignore_regex = f"[{re.escape(''.join(CHARS_TO_IGNORE))}]"
    batch["sentence"] = re.sub(chars_to_ignore_regex, "", batch["sentence"]).upper().strip()
    return batch


def fun_ka(batch):
    chars_to_ignore_regex = '[\\\\\\\\,\\\\\\\\?\\\\\\\\.\\\\\\\\!\\\\\\\\-\\\\\\\\;\\\\\\\\:\\\\\\\\"\\\\\\\\“]'
    batch["sentence"] = re.sub(chars_to_ignore_regex, '', batch["sentence"]).lower()
    return batch


def fun_ky(batch):
    chars_to_ignore = [",", "?", ".", "!", "-", ";", ":", "—", "–", "”"]
    chars_to_ignore_regex = f'[{"".join(chars_to_ignore)}]'
    batch["sentence"] = re.sub(chars_to_ignore_regex, '', batch["sentence"]).lower()
    return batch


def fun_lg(batch):
    import unidecode
    chars_to_ignore_regex = '[\[\],?.!;:%"“”(){}‟ˮʺ″«»/…‽�–]'
    batch["sentence"] = re.sub(r'(\w)[‘’´`](\w)', r"\1'\2", batch["sentence"])
    batch["sentence"] = re.sub(chars_to_ignore_regex, "", batch["sentence"]).lower().strip()
    batch["sentence"] = re.sub(r"(-|' | '|  +)", " ", batch["sentence"])
    batch["sentence"] = unidecode.unidecode(batch["sentence"]).strip()
    return batch


def fun_lt(batch):
    chars_to_ignore_regex = '[\,\?\.\!\-\;\:\"\“\%\‘\”\�]'
    batch["sentence"] = re.sub(chars_to_ignore_regex, '', batch["sentence"]).lower()
    return batch


def fun_lv(batch):
    sent = batch["sentence"].lower()
    # replace non-alpha characters with space
    sent = "".join(ch if ch.isalpha() else " " for ch in sent)
    # remove repeated spaces
    sent = " ".join(sent.split())
    batch["sentence"] = sent
    return batch


def fun_mn(batch):
    sent = batch["sentence"].lower()
    # replace non-alpha characters with space
    sent = "".join(ch if ch.isalpha() else " " for ch in sent)
    # remove repeated spaces
    sent = " ".join(sent.split())
    batch["sentence"] = sent
    return batch


def fun_mt(batch):
    chars_to_ignore_regex = '[\\,\\?\\.\\!\\-\\;\\:\\"\\“\\%\\‘\\”\\�]'
    batch["sentence"] = re.sub(chars_to_ignore_regex, '', batch["sentence"]).lower() + " "
    return batch


def fun_nl(batch):
    chars_to_ignore_regex = '[\,\?\.\!\-\;\:\"\'\“\%\‘\”]'
    batch["sentence"] = re.sub(chars_to_ignore_regex, '', batch["sentence"]).lower()
    return batch


def fun_or(batch):
    chars_to_ignore_regex = '[\,\?\.\!\-\;\:\"\“\%\‘\”\�\–\…\'\_\’\।\|]'
    batch["sentence"] = re.sub(chars_to_ignore_regex, '', batch["sentence"]).lower()
    return batch


def fun_pa_IN(batch):
    chars_to_ignore_regex = '[\,\?\.\!\-\;\:\"\“\%\‘\”\�\’\–\(\)]'
    batch["sentence"] = re.sub(chars_to_ignore_regex, '', batch["sentence"]).lower()
    return batch


def fun_pl(batch):
    chars_to_ignore_regex = '[\—\…\,\?\.\!\-\;\:\"\“\„\%\‘\”\�\«\»\'\’]'
    batch["sentence"] = re.sub(chars_to_ignore_regex, '', batch["sentence"]).lower()
    return batch


def fun_pt(batch):
    CHARS_TO_IGNORE = [",", "?", "¿", ".", "!", "¡", ";", ":", '""', "%", '"', "�", "ʿ", "·", "჻", "~", "՞",
                       "؟", "،", "।", "॥", "«", "»", "„", "“", "”", "「", "」", "‘", "’", "《", "》", "(", ")", "[", "]",
                       "=", "`", "_", "+", "<", ">", "…", "–", "°", "´", "ʾ", "‹", "›", "©", "®", "—", "→", "。"]
    chars_to_ignore_regex = f"[{re.escape(''.join(CHARS_TO_IGNORE))}]"
    batch["sentence"] = re.sub(chars_to_ignore_regex, "", batch["sentence"]).upper()
    return batch


def fun_rm_sursilv(batch):
    chars_to_ignore_regex = '[\\,\\?\\.\\!\\-\\;\\:\\"\\“\\%\\‘\\”\\�\\…\\«\\»\\–]'
    batch["sentence"] = re.sub(chars_to_ignore_regex, '', batch["sentence"]).lower()
    return batch


def fun_rm_vallader(batch):
    chars_to_ignore_regex = '[\,\?\.\!\-\;\:\"\“\%\”\„\–\…\«\»]'
    batch["sentence"] = re.sub('’ ', ' ', batch["sentence"])
    batch["sentence"] = re.sub(' ‘', ' ', batch["sentence"])
    batch["sentence"] = re.sub('’|‘', '\'', batch["sentence"])
    batch["sentence"] = re.sub(chars_to_ignore_regex, '', batch["sentence"]).lower()
    return batch


def fun_ro(batch):
    sent = batch["sentence"].lower()
    # replace non-alpha characters with space
    sent = "".join(ch if ch.isalpha() else " " for ch in sent)
    # remove repeated spaces
    sent = " ".join(sent.split())
    batch["sentence"] = sent
    return batch


def fun_ru(batch):
    sent = batch["sentence"].lower()
    # these letters are considered equivalent in written Russian
    sent = sent.replace('ё', 'е')
    # replace non-alpha characters with space
    sent = "".join(ch if ch.isalpha() else " " for ch in sent)
    # remove repeated spaces
    sent = " ".join(sent.split())
    batch["sentence"] = sent
    return batch


def fun_sah(batch):
    sent = batch["sentence"].lower()
    # replace non-alpha characters with space
    sent = "".join(ch if ch.isalpha() else " " for ch in sent)
    # remove repeated spaces
    sent = " ".join(sent.split())
    batch["sentence"] = sent
    return batch


def fun_sl(batch):
    sent = batch["sentence"].lower()
    # replace non-alpha characters with space
    sent = "".join(ch if ch.isalpha() else " " for ch in sent)
    # remove repeated spaces
    sent = " ".join(sent.split())
    batch["sentence"] = sent
    return batch


def fun_sv_SE(batch):
    chars_to_ignore_regex = '[,?.!\\-;:"“]'
    batch["sentence"] = re.sub(chars_to_ignore_regex, '', batch["sentence"]).lower()
    return batch


def fun_ta(batch):
    chars_to_ignore_regex = '[\,\?\.\!\-\;\:\"\“\%\‘\”\�\’\–\(\)]'
    batch["sentence"] = re.sub(chars_to_ignore_regex, '', batch["sentence"]).lower()
    return batch


def fun_th(batch):
    chars_to_ignore_regex = '[\,\?\.\!\-\;\:\"\â€œ]'
    batch["sentence"] = re.sub(chars_to_ignore_regex, '', batch["sentence"]).lower()
    return batch


def fun_tr(batch):
    chars_to_ignore_regex = '[\,\?\.\!\-\;\'\:\"\“\%\‘\”\�]'
    batch["sentence"] = re.sub(chars_to_ignore_regex, '', batch["sentence"]).lower()
    return batch


def fun_tt(batch):
    sent = batch["sentence"].lower()
    # 'ё' is equivalent to 'е'
    sent = sent.replace('ё', 'е')
    # replace non-alpha characters with space
    sent = "".join(ch if ch.isalpha() else " " for ch in sent)
    # remove repeated spaces
    batch["sentence"] = " ".join(sent.split())
    return batch


def fun_uk(batch):
    chars_to_ignore = [",", "?", ".", "!", "-", ";", ":", '""', "%", "'", '"', "�", '«', '»', '—', '…', '(', ')', '*',
                       '”', '“']
    chars_to_ignore_regex = f'[{"".join(chars_to_ignore)}]'
    batch["sentence"] = re.sub(re.compile("['`]"), '’', batch['sentence'])
    batch["sentence"] = re.sub(re.compile(chars_to_ignore_regex), '', batch["sentence"]).lower().strip()
    batch["sentence"] = re.sub(re.compile('i'), 'і', batch['sentence'])
    batch["sentence"] = re.sub(re.compile('o'), 'о', batch['sentence'])
    batch["sentence"] = re.sub(re.compile('a'), 'а', batch['sentence'])
    batch["sentence"] = re.sub(re.compile('ы'), 'и', batch['sentence'])
    batch["sentence"] = re.sub(re.compile("–"), '', batch['sentence'])
    batch['sentence'] = re.sub('  ', ' ', batch['sentence'])
    return batch


def fun_vi(batch):
    chars_to_ignore_regex = '[\\\+\@\ǀ\,\?\.\!\-\;\:\"\“\%\‘\”\�]'
    batch["sentence"] = re.sub(chars_to_ignore_regex, '', batch["sentence"]).lower()
    return batch


def fun_zh_CN(batch):
    chars_to_ignore_regex = r"[¥•＂＃＄％＆＇（）＊＋，－／：；＜＝＞＠［＼］＾＿｀｛｜｝～｟｠｢｣､　、〃〈〉《》「」『』【】〔〕〖〗〘〙〚〛〜〝〞〟〰〾〿–—‘’‛“”„‟…‧﹏﹑﹔·'℃°•·．﹑︰〈〉─《﹖﹣﹂﹁﹔！？｡。＂＃＄％＆＇（）＊＋，﹐－／：；＜＝＞＠［＼］＾＿｀｛｜｝～｟｠｢｣､、〃》「」『』【】〔〕〖〗〘〙〚〛〜〝〞〟〰〾〿–—‘’‛“”„‟…‧﹏.．!\"#$%&()*+,\-.\:;<=>?@\[\]\\\/^_`{|}~]"
    batch["sentence"] = re.sub(chars_to_ignore_regex, '', batch["sentence"]).lower()
    return batch


def fun_zh_HK(batch):
    chars_to_ignore_regex = r"[¥•＂＃＄％＆＇（）＊＋，－／：；＜＝＞＠［＼］＾＿｀｛｜｝～｟｠｢｣､　、〃〈〉《》「」『』【】〔〕〖〗〘〙〚〛〜〝〞〟〰〾〿–—‘’‛“”„‟…‧﹏﹑﹔·'℃°•·．﹑︰〈〉─《﹖﹣﹂﹁﹔！？｡。＂＃＄％＆＇（）＊＋，﹐－／：；＜＝＞＠［＼］＾＿｀｛｜｝～｟｠｢｣､、〃》「」『』【】〔〕〖〗〘〙〚〛〜〝〞〟〰〾〿–—‘’‛“”„‟…‧﹏.．!\"#$%&()*+,\-.\:;<=>?@\[\]\\\/^_`{|}~]"
    batch["sentence"] = re.sub(chars_to_ignore_regex, '', batch["sentence"]).lower()
    return batch


def fun_zh_TW(batch):
    chars_to_ignore_regex = r"[¥•＂＃＄％＆＇（）＊＋，－／：；＜＝＞＠［＼］＾＿｀｛｜｝～｟｠｢｣､　、〃〈〉《》「」『』【】〔〕〖〗〘〙〚〛〜〝〞〟〰〾〿–—‘’‛“”„‟…‧﹏﹑﹔·'℃°•·．﹑︰〈〉─《﹖﹣﹂﹁﹔！？｡。＂＃＄％＆＇（）＊＋，﹐－／：；＜＝＞＠［＼］＾＿｀｛｜｝～｟｠｢｣､、〃》「」『』【】〔〕〖〗〘〙〚〛〜〝〞〟〰〾〿–—‘’‛“”„‟…‧﹏.．!\"#$%&()*+,\-.\:;<=>?@\[\]\\\/^_`{|}~]"
    batch["sentence"] = re.sub(chars_to_ignore_regex, '', batch["sentence"]).lower()
    return batch



import nlp2

from asrp import Code2Speech


def hifigan_hubert_layer6_code100():
    # https://github.com/facebookresearch/fairseq/blob/ust/examples/speech_to_speech/docs/direct_s2st_discrete_units.md
    nlp2.download_file(
        'https://dl.fbaipublicfiles.com/fairseq/speech_to_speech/vocoder/code_hifigan/hubert_base_100_lj/g_00500000',
        './', 'hifigan_hubert_layer6_code100_g_00500000')
    nlp2.download_file(
        'https://dl.fbaipublicfiles.com/fairseq/speech_to_speech/vocoder/code_hifigan/hubert_base_100_lj/config.json',
        './', 'hifigan_hubert_layer6_code100_config.json')
    cs = Code2Speech(tts_checkpoint='./hifigan_hubert_layer6_code100_g_00500000',
                     model_cfg=nlp2.read_json('./hifigan_hubert_layer6_code100_config.json'), vocoder='hifigan')
    return cs



from pathlib import Path

import torch
import torchaudio

from .denoiser.pretrained import master64
from .denoiser.utils import cal_snr
from .utility import write, collate_fn_padd


class SpeechEnhancer:
    def __init__(self, device='cpu', dry_wet=0.01, sampling_rate=16000, chunk_sec=30, max_batch=3):
        torchaudio.set_audio_backend("sox_io")  # switch backend
        self.device = device
        self.dry_wet = dry_wet
        self.sampling_rate = sampling_rate
        self.chunk_sec = chunk_sec
        self.max_batch = max_batch
        self.chunk_length = sampling_rate * chunk_sec

        self.enhance_model = master64()
        self.enhance_model = self.enhance_model.to(self.device)
        self.enhance_model.eval()

    def __call__(self,
                 filepath='',
                 input_values=[],
                 result_path=''):

        if len(filepath) > 0:
            out, sr = torchaudio.load(filepath)
            out = out.mean(0).unsqueeze(0)
        else:
            out = input_values

        # split audio into chunks
        chunks = list(torch.split(out, self.chunk_length, dim=1))
        if chunks[-1].shape[-1] < self.sampling_rate:
            concat_index = -2 if len(chunks) >= 2 else 0
            chunks[concat_index] = torch.cat(chunks[-2:], dim=-1)
            chunks = chunks[:concat_index + 1]

        batch_data = []
        cache_batch = []
        for c in chunks:
            if len(cache_batch) >= self.max_batch:
                batch_data.append(cache_batch)
                cache_batch = []
            cache_batch.append(c)
        if len(cache_batch) > 0:
            batch_data.append(cache_batch)

        enhance_result = []
        for bd in batch_data:
            batch, lengths, masks = collate_fn_padd([i[0] for i in bd], self.device)
            estimate = (1 - self.dry_wet) * self.enhance_model(batch).squeeze(1) + self.dry_wet * batch
            enhance_result.append(torch.masked_select(estimate, masks).detach().cpu())

        denoise = torch.cat(enhance_result, dim=-1).unsqueeze(0)

        p = Path(filepath)
        write(denoise, str(Path(p.parent, f"{p.stem}_enhanced{p.suffix}")), sr)
        snr = cal_snr(out, denoise)
        snr = snr.cpu().detach().numpy()[0]
        return snr



import torch
from math import sqrt
import torch.nn.functional as F
from torch.autograd import Variable
from torch import nn


class LinearNorm(torch.nn.Module):
    def __init__(self, in_dim, out_dim, bias=True, w_init_gain='linear'):
        super(LinearNorm, self).__init__()
        self.linear_layer = torch.nn.Linear(in_dim, out_dim, bias=bias)

        torch.nn.init.xavier_uniform_(
            self.linear_layer.weight,
            gain=torch.nn.init.calculate_gain(w_init_gain))

    def forward(self, x):
        return self.linear_layer(x)


class ConvNorm(torch.nn.Module):
    def __init__(self, in_channels, out_channels, kernel_size=1, stride=1,
                 padding=None, dilation=1, bias=True, w_init_gain='linear'):
        super(ConvNorm, self).__init__()
        if padding is None:
            assert (kernel_size % 2 == 1)
            padding = int(dilation * (kernel_size - 1) / 2)

        self.conv = torch.nn.Conv1d(in_channels, out_channels,
                                    kernel_size=kernel_size, stride=stride,
                                    padding=padding, dilation=dilation,
                                    bias=bias)

        torch.nn.init.xavier_uniform_(
            self.conv.weight, gain=torch.nn.init.calculate_gain(w_init_gain))

    def forward(self, signal):
        conv_signal = self.conv(signal)
        return conv_signal


class Encoder(nn.Module):
    """Encoder module:
        - Three 1-d convolution banks
        - Bidirectional LSTM
    """

    def __init__(self, hparams):
        super(Encoder, self).__init__()

        convolutions = []
        for _ in range(hparams.encoder_n_convolutions):
            conv_layer = nn.Sequential(
                ConvNorm(hparams.encoder_embedding_dim,
                         hparams.encoder_embedding_dim,
                         kernel_size=hparams.encoder_kernel_size, stride=1,
                         padding=int((hparams.encoder_kernel_size - 1) / 2),
                         dilation=1, w_init_gain='relu'),
                nn.BatchNorm1d(hparams.encoder_embedding_dim))
            convolutions.append(conv_layer)
        self.convolutions = nn.ModuleList(convolutions)

        self.lstm = nn.LSTM(hparams.encoder_embedding_dim,
                            int(hparams.encoder_embedding_dim / 2), 1,
                            batch_first=True, bidirectional=True)

    def forward(self, x, input_lengths):
        for conv in self.convolutions:
            x = F.dropout(F.relu(conv(x)), 0.5, self.training)

        x = x.transpose(1, 2)

        # pytorch tensor are not reversible, hence the conversion
        input_lengths = input_lengths.cpu().numpy()
        x = nn.utils.rnn.pack_padded_sequence(
            x, input_lengths, batch_first=True)

        self.lstm.flatten_parameters()
        outputs, _ = self.lstm(x)

        outputs, _ = nn.utils.rnn.pad_packed_sequence(
            outputs, batch_first=True)

        return outputs

    def inference(self, x):
        for conv in self.convolutions:
            x = F.dropout(F.relu(conv(x)), 0.5, self.training)

        x = x.transpose(1, 2)

        self.lstm.flatten_parameters()
        outputs, _ = self.lstm(x)

        return outputs


class Decoder(nn.Module):
    def __init__(self, hparams):
        super(Decoder, self).__init__()
        self.n_mel_channels = hparams.n_mel_channels
        self.n_frames_per_step = hparams.n_frames_per_step
        self.encoder_embedding_dim = hparams.encoder_embedding_dim
        self.obs_dim = hparams.obs_dim
        self.lat_dim = hparams.lat_dim
        self.attention_rnn_dim = hparams.attention_rnn_dim
        self.decoder_rnn_dim = hparams.decoder_rnn_dim
        self.prenet_dim = hparams.prenet_dim
        self.max_decoder_steps = hparams.max_decoder_steps
        self.gate_threshold = hparams.gate_threshold
        self.p_attention_dropout = hparams.p_attention_dropout
        self.p_decoder_dropout = hparams.p_decoder_dropout

        self.prenet = Prenet(
            hparams.n_mel_channels * hparams.n_frames_per_step,
            [hparams.prenet_dim, hparams.prenet_dim])

        self.attention_rnn = nn.LSTMCell(
            hparams.prenet_dim + hparams.encoder_embedding_dim,
            hparams.attention_rnn_dim)

        self.attention_layer = Attention(
            hparams.attention_rnn_dim, hparams.encoder_embedding_dim,
            hparams.attention_dim, hparams.attention_location_n_filters,
            hparams.attention_location_kernel_size)

        encoder_tot_dim = (hparams.encoder_embedding_dim + \
                           hparams.lat_dim + hparams.obs_dim)
        self.decoder_rnn = nn.LSTMCell(
            hparams.attention_rnn_dim + encoder_tot_dim,
            hparams.decoder_rnn_dim, 1)

        self.linear_projection = LinearNorm(
            hparams.decoder_rnn_dim + encoder_tot_dim,
            hparams.n_mel_channels * hparams.n_frames_per_step)

        self.gate_layer = LinearNorm(
            hparams.decoder_rnn_dim + encoder_tot_dim, 1,
            bias=True, w_init_gain='sigmoid')

    def get_go_frame(self, memory):
        """ Gets all zeros frames to use as first decoder input
        PARAMS
        ------
        memory: decoder outputs
        RETURNS
        -------
        decoder_input: all zeros frames
        """
        B = memory.size(0)
        decoder_input = Variable(memory.data.new(
            B, self.n_mel_channels * self.n_frames_per_step).zero_())
        return decoder_input

    def initialize_decoder_states(self, memory, obs_and_lat, mask):
        """ Initializes attention rnn states, decoder rnn states, attention
        weights, attention cumulative weights, attention context, stores memory
        and stores processed memory
        PARAMS
        ------
        memory: Encoder outputs
        obs_and_lat: Observed and latent attribute embeddings
        mask: Mask for padded data if training, expects None for inference
        """
        B = memory.size(0)
        MAX_TIME = memory.size(1)

        self.attention_hidden = Variable(memory.data.new(
            B, self.attention_rnn_dim).zero_())
        self.attention_cell = Variable(memory.data.new(
            B, self.attention_rnn_dim).zero_())

        self.decoder_hidden = Variable(memory.data.new(
            B, self.decoder_rnn_dim).zero_())
        self.decoder_cell = Variable(memory.data.new(
            B, self.decoder_rnn_dim).zero_())

        self.attention_weights = Variable(memory.data.new(
            B, MAX_TIME).zero_())
        self.attention_weights_cum = Variable(memory.data.new(
            B, MAX_TIME).zero_())
        self.attention_context = Variable(memory.data.new(
            B, self.encoder_embedding_dim).zero_())

        self.memory = memory
        self.processed_memory = self.attention_layer.memory_layer(memory)
        self.obs_and_lat = obs_and_lat
        self.mask = mask

    def parse_decoder_inputs(self, decoder_inputs):
        """ Prepares decoder inputs, i.e. mel outputs
        PARAMS
        ------
        decoder_inputs: inputs used for teacher-forced training, i.e. mel-specs
        RETURNS
        -------
        inputs: processed decoder inputs
        """
        # (B, n_mel_channels, T_out) -> (B, T_out, n_mel_channels)
        decoder_inputs = decoder_inputs.transpose(1, 2)
        decoder_inputs = decoder_inputs.view(
            decoder_inputs.size(0),
            int(decoder_inputs.size(1) / self.n_frames_per_step), -1)
        # (B, T_out, n_mel_channels) -> (T_out, B, n_mel_channels)
        decoder_inputs = decoder_inputs.transpose(0, 1)
        return decoder_inputs

    def parse_decoder_outputs(self, mel_outputs, gate_outputs, alignments):
        """ Prepares decoder outputs for output
        PARAMS
        ------
        mel_outputs:
        gate_outputs: gate output energies
        alignments:
        RETURNS
        -------
        mel_outputs:
        gate_outpust: gate output energies
        alignments:
        """
        # (T_out, B) -> (B, T_out)
        alignments = torch.stack(alignments).transpose(0, 1)
        # (T_out, B) -> (B, T_out)
        gate_outputs = torch.stack(gate_outputs).transpose(0, 1)
        gate_outputs = gate_outputs.contiguous()
        # (T_out, B, n_mel_channels) -> (B, T_out, n_mel_channels)
        mel_outputs = torch.stack(mel_outputs).transpose(0, 1).contiguous()
        # decouple frames per step
        mel_outputs = mel_outputs.view(
            mel_outputs.size(0), -1, self.n_mel_channels)
        # (B, T_out, n_mel_channels) -> (B, n_mel_channels, T_out)
        mel_outputs = mel_outputs.transpose(1, 2)

        return mel_outputs, gate_outputs, alignments

    def decode(self, decoder_input):
        """ Decoder step using stored states, attention and memory
        PARAMS
        ------
        decoder_input: previous mel output
        RETURNS
        -------
        mel_output:
        gate_output: gate output energies
        attention_weights:
        """
        cell_input = torch.cat((decoder_input, self.attention_context), -1)
        self.attention_hidden, self.attention_cell = self.attention_rnn(
            cell_input, (self.attention_hidden, self.attention_cell))
        self.attention_hidden = F.dropout(
            self.attention_hidden, self.p_attention_dropout, self.training)

        attention_weights_cat = torch.cat(
            (self.attention_weights.unsqueeze(1),
             self.attention_weights_cum.unsqueeze(1)), dim=1)
        self.attention_context, self.attention_weights = self.attention_layer(
            self.attention_hidden, self.memory, self.processed_memory,
            attention_weights_cat, self.mask)

        self.attention_weights_cum += self.attention_weights
        decoder_input = torch.cat(
            (self.attention_hidden, self.attention_context), -1)
        if self.obs_and_lat is not None:
            decoder_input = torch.cat((decoder_input, self.obs_and_lat), -1)
        self.decoder_hidden, self.decoder_cell = self.decoder_rnn(
            decoder_input, (self.decoder_hidden, self.decoder_cell))
        self.decoder_hidden = F.dropout(
            self.decoder_hidden, self.p_decoder_dropout, self.training)

        decoder_hidden_attention_context = torch.cat(
            (self.decoder_hidden, self.attention_context), dim=1)
        if self.obs_and_lat is not None:
            decoder_hidden_attention_context = torch.cat(
                (decoder_hidden_attention_context, self.obs_and_lat), dim=1)
        decoder_output = self.linear_projection(
            decoder_hidden_attention_context)

        gate_prediction = self.gate_layer(decoder_hidden_attention_context)
        return decoder_output, gate_prediction, self.attention_weights

    def forward(self, memory, obs_and_lat, decoder_inputs, memory_lengths):
        """ Decoder forward pass for training
        PARAMS
        ------
        memory: Encoder outputs
        obs_and_lat: Observed and latent attribute embeddings
        decoder_inputs: Decoder inputs for teacher forcing. i.e. mel-specs
        memory_lengths: Encoder output lengths for attention masking.
        RETURNS
        -------
        mel_outputs: mel outputs from the decoder
        gate_outputs: gate outputs from the decoder
        alignments: sequence of attention weights from the decoder
        """

        decoder_input = self.get_go_frame(memory).unsqueeze(0)
        decoder_inputs = self.parse_decoder_inputs(decoder_inputs)
        decoder_inputs = torch.cat((decoder_input, decoder_inputs), dim=0)
        decoder_inputs = self.prenet(decoder_inputs)

        self.initialize_decoder_states(
            memory, obs_and_lat, mask=~get_mask_from_lengths(memory_lengths))

        mel_outputs, gate_outputs, alignments = [], [], []
        while len(mel_outputs) < decoder_inputs.size(0) - 1:
            decoder_input = decoder_inputs[len(mel_outputs)]
            mel_output, gate_output, attention_weights = self.decode(
                decoder_input)
            mel_outputs += [mel_output.squeeze(1)]
            gate_outputs += [gate_output.squeeze()]
            alignments += [attention_weights]

        mel_outputs, gate_outputs, alignments = self.parse_decoder_outputs(
            mel_outputs, gate_outputs, alignments)

        return mel_outputs, gate_outputs, alignments

    def inference(self, memory, obs_and_lat, ret_has_eos=False):
        """ Decoder inference
        PARAMS
        ------
        memory: Encoder outputs
        obs_and_lat: Observed and latent attribute embeddings
        RETURNS
        -------
        mel_outputs: mel outputs from the decoder
        gate_outputs: gate outputs from the decoder
        alignments: sequence of attention weights from the decoder
        """
        decoder_input = self.get_go_frame(memory)

        self.initialize_decoder_states(memory, obs_and_lat, mask=None)

        mel_outputs, gate_outputs, alignments = [], [], []
        has_eos = False
        while True:
            decoder_input = self.prenet(decoder_input)
            mel_output, gate_output, alignment = self.decode(decoder_input)

            mel_outputs += [mel_output.squeeze(1)]
            gate_outputs += [gate_output]
            alignments += [alignment]

            if torch.sigmoid(gate_output.data) > self.gate_threshold:
                has_eos = True
                break
            elif len(mel_outputs) == self.max_decoder_steps:
                # print("Warning! Reached max decoder steps")
                break

            decoder_input = mel_output

        mel_outputs, gate_outputs, alignments = self.parse_decoder_outputs(
            mel_outputs, gate_outputs, alignments)

        if ret_has_eos:
            return mel_outputs, gate_outputs, alignments, has_eos
        else:
            return mel_outputs, gate_outputs, alignments


class Attention(nn.Module):
    def __init__(self, attention_rnn_dim, embedding_dim, attention_dim,
                 attention_location_n_filters, attention_location_kernel_size):
        super(Attention, self).__init__()
        self.query_layer = LinearNorm(attention_rnn_dim, attention_dim,
                                      bias=False, w_init_gain='tanh')
        self.memory_layer = LinearNorm(embedding_dim, attention_dim, bias=False,
                                       w_init_gain='tanh')
        self.v = LinearNorm(attention_dim, 1, bias=False)
        self.location_layer = LocationLayer(attention_location_n_filters,
                                            attention_location_kernel_size,
                                            attention_dim)
        self.score_mask_value = -float("inf")

    def get_alignment_energies(self, query, processed_memory,
                               attention_weights_cat):
        """
        PARAMS
        ------
        query: decoder output (batch, n_mel_channels * n_frames_per_step)
        processed_memory: processed encoder outputs (B, T_in, attention_dim)
        attention_weights_cat: cumulative and prev. att weights (B, 2, max_time)
        RETURNS
        -------
        alignment (batch, max_time)
        """

        processed_query = self.query_layer(query.unsqueeze(1))
        processed_attention_weights = self.location_layer(attention_weights_cat)
        energies = self.v(torch.tanh(
            processed_query + processed_attention_weights + processed_memory))

        energies = energies.squeeze(-1)
        return energies

    def forward(self, attention_hidden_state, memory, processed_memory,
                attention_weights_cat, mask):
        """
        PARAMS
        ------
        attention_hidden_state: attention rnn last output
        memory: encoder outputs
        processed_memory: processed encoder outputs
        attention_weights_cat: previous and cummulative attention weights
        mask: binary mask for padded data
        """
        alignment = self.get_alignment_energies(
            attention_hidden_state, processed_memory, attention_weights_cat)

        if mask is not None:
            alignment.data.masked_fill_(mask, self.score_mask_value)

        attention_weights = F.softmax(alignment, dim=1)
        attention_context = torch.bmm(attention_weights.unsqueeze(1), memory)
        attention_context = attention_context.squeeze(1)

        return attention_context, attention_weights


class Prenet(nn.Module):
    def __init__(self, in_dim, sizes):
        super(Prenet, self).__init__()
        in_sizes = [in_dim] + sizes[:-1]
        self.layers = nn.ModuleList(
            [LinearNorm(in_size, out_size, bias=False)
             for (in_size, out_size) in zip(in_sizes, sizes)])

    def forward(self, x):
        for linear in self.layers:
            x = F.dropout(F.relu(linear(x)), p=0.5, training=True)
        return x


class Postnet(nn.Module):
    """Postnet
        - Five 1-d convolution with 512 channels and kernel size 5
    """

    def __init__(self, hparams):
        super(Postnet, self).__init__()
        self.convolutions = nn.ModuleList()

        self.convolutions.append(
            nn.Sequential(
                ConvNorm(hparams.n_mel_channels, hparams.postnet_embedding_dim,
                         kernel_size=hparams.postnet_kernel_size, stride=1,
                         padding=int((hparams.postnet_kernel_size - 1) / 2),
                         dilation=1, w_init_gain='tanh'),
                nn.BatchNorm1d(hparams.postnet_embedding_dim))
        )

        for i in range(1, hparams.postnet_n_convolutions - 1):
            self.convolutions.append(
                nn.Sequential(
                    ConvNorm(hparams.postnet_embedding_dim,
                             hparams.postnet_embedding_dim,
                             kernel_size=hparams.postnet_kernel_size, stride=1,
                             padding=int((hparams.postnet_kernel_size - 1) / 2),
                             dilation=1, w_init_gain='tanh'),
                    nn.BatchNorm1d(hparams.postnet_embedding_dim))
            )

        self.convolutions.append(
            nn.Sequential(
                ConvNorm(hparams.postnet_embedding_dim, hparams.n_mel_channels,
                         kernel_size=hparams.postnet_kernel_size, stride=1,
                         padding=int((hparams.postnet_kernel_size - 1) / 2),
                         dilation=1, w_init_gain='linear'),
                nn.BatchNorm1d(hparams.n_mel_channels))
        )

    def forward(self, x):
        for i in range(len(self.convolutions) - 1):
            x = F.dropout(torch.tanh(self.convolutions[i](x)), 0.5, self.training)
        x = F.dropout(self.convolutions[-1](x), 0.5, self.training)

        return x


class LocationLayer(nn.Module):
    def __init__(self, attention_n_filters, attention_kernel_size,
                 attention_dim):
        super(LocationLayer, self).__init__()
        padding = int((attention_kernel_size - 1) / 2)
        self.location_conv = ConvNorm(2, attention_n_filters,
                                      kernel_size=attention_kernel_size,
                                      padding=padding, bias=False, stride=1,
                                      dilation=1)
        self.location_dense = LinearNorm(attention_n_filters, attention_dim,
                                         bias=False, w_init_gain='tanh')

    def forward(self, attention_weights_cat):
        processed_attention = self.location_conv(attention_weights_cat)
        processed_attention = processed_attention.transpose(1, 2)
        processed_attention = self.location_dense(processed_attention)
        return processed_attention


class Tacotron2(nn.Module):
    def __init__(self, hparams):
        super(Tacotron2, self).__init__()
        self.mask_padding = hparams.mask_padding
        self.fp16_run = hparams.fp16_run
        self.n_mel_channels = hparams.n_mel_channels
        self.n_frames_per_step = hparams.n_frames_per_step

        # initialize text encoder embedding
        self.embedding = nn.Embedding(
            hparams.n_symbols, hparams.symbols_embedding_dim)
        std = sqrt(2.0 / (hparams.n_symbols + hparams.symbols_embedding_dim))
        val = sqrt(3.0) * std  # uniform bounds for std
        self.embedding.weight.data.uniform_(-val, val)

        # initialize observed attribute embedding
        self.obs_embedding = None
        if hparams.obs_dim > 0:
            self.obs_embedding = nn.Embedding(
                hparams.obs_n_class, hparams.obs_dim)
            std = sqrt(2.0 / (hparams.obs_n_class + hparams.obs_dim))
            val = sqrt(3.0) * std  # uniform bounds for std
            self.obs_embedding.weight.data.uniform_(-val, val)

        self.encoder = Encoder(hparams)
        self.decoder = Decoder(hparams)
        self.postnet = Postnet(hparams)

        self.lat_encoder = None
        if hparams.lat_dim > 0:
            self.lat_encoder = AudioEncoder(hparams)

    def parse_batch(self, batch):
        (text_padded, input_lengths, obs_labels,
         mel_padded, gate_padded, output_lengths) = batch
        text_padded = to_gpu(text_padded).long()
        input_lengths = to_gpu(input_lengths).long()
        obs_labels = to_gpu(obs_labels).long()
        max_len = torch.max(input_lengths.data).item()
        mel_padded = to_gpu(mel_padded).float()
        gate_padded = to_gpu(gate_padded).float()
        output_lengths = to_gpu(output_lengths).long()

        return (
            (text_padded, input_lengths, obs_labels,
             mel_padded, max_len, output_lengths),
            (mel_padded, gate_padded))

    def parse_output(self, outputs, output_lengths=None):
        if self.mask_padding and output_lengths is not None:
            mask = ~get_mask_from_lengths(output_lengths)
            mask = mask.expand(self.n_mel_channels, mask.size(0), mask.size(1))
            mask = mask.permute(1, 0, 2)

            outputs[0].data.masked_fill_(mask, 0.0)
            outputs[1].data.masked_fill_(mask, 0.0)
            outputs[2].data.masked_fill_(mask[:, 0, :], 1e3)  # gate energies

        return outputs

    def forward(self, inputs):
        (text_inputs, text_lengths, obs_labels,
         mels, max_len, output_lengths) = inputs
        text_lengths, output_lengths = text_lengths.data, output_lengths.data

        embedded_inputs = self.embedding(text_inputs).transpose(1, 2)

        encoder_outputs = self.encoder(embedded_inputs, text_lengths)

        obs = None
        if self.obs_embedding is not None:
            obs = self.obs_embedding(obs_labels)

        lat, lat_mu, lat_logvar = None, None, None
        if self.lat_encoder is not None:
            (lat, lat_mu, lat_logvar) = self.lat_encoder(mels, output_lengths)

        obs_and_lat = [x for x in [obs, lat] if x is not None]
        if bool(obs_and_lat):
            obs_and_lat = torch.cat(obs_and_lat, dim=-1)
        else:
            obs_and_lat = None

        mel_outputs, gate_outputs, alignments = self.decoder(
            encoder_outputs, obs_and_lat, mels, memory_lengths=text_lengths)

        mel_outputs_postnet = self.postnet(mel_outputs)
        mel_outputs_postnet = mel_outputs + mel_outputs_postnet

        return self.parse_output(
            [mel_outputs, mel_outputs_postnet, gate_outputs, alignments,
             lat_mu, lat_logvar],
            output_lengths)

    def inference(self, inputs, obs_labels=None, lat=None, ret_has_eos=False):
        embedded_inputs = self.embedding(inputs).transpose(1, 2)
        encoder_outputs = self.encoder.inference(embedded_inputs)

        if obs_labels is None:
            obs_labels = torch.LongTensor(len(inputs))
            obs_labels = obs_labels.to(inputs.device).zero_()

        obs = None
        if self.obs_embedding is not None:
            obs = self.obs_embedding(obs_labels)

        if self.lat_encoder is not None:
            if lat is None:
                lat = torch.FloatTensor(len(inputs), self.lat_encoder.lat_dim)
                lat = lat.to(inputs.device).zero_().type(encoder_outputs.type())

        obs_and_lat = [x for x in [obs, lat] if x is not None]
        if bool(obs_and_lat):
            obs_and_lat = torch.cat(obs_and_lat, dim=-1)
        else:
            obs_and_lat = None

        mel_outputs, gate_outputs, alignments, has_eos = self.decoder.inference(
            encoder_outputs, obs_and_lat, ret_has_eos=True)

        mel_outputs_postnet = self.postnet(mel_outputs)
        mel_outputs_postnet = mel_outputs + mel_outputs_postnet

        outputs = self.parse_output(
            [mel_outputs, mel_outputs_postnet, gate_outputs, alignments])

        if ret_has_eos:
            return outputs + [has_eos]
        else:
            return outputs



# Copyright (c) Facebook, Inc. and its affiliates.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.
# author: adefossez

import logging

import torch.hub

from .demucs import Demucs
from .utils import deserialize_model

logger = logging.getLogger(__name__)
ROOT = "https://dl.fbaipublicfiles.com/adiyoss/denoiser/"
DNS_48_URL = ROOT + "dns48-11decc9d8e3f0998.th"
DNS_64_URL = ROOT + "dns64-a7761ff99a7d5bb6.th"
MASTER_64_URL = ROOT + "master64-8a5dfb4bb92753dd.th"


def _demucs(pretrained, url, **kwargs):
    model = Demucs(**kwargs)
    if pretrained:
        state_dict = torch.hub.load_state_dict_from_url(url, map_location='cpu')
        model.load_state_dict(state_dict)
    return model


def dns48(pretrained=True):
    return _demucs(pretrained, DNS_48_URL, hidden=48)


def dns64(pretrained=True):
    return _demucs(pretrained, DNS_64_URL, hidden=64)


def master64(pretrained=True):
    return _demucs(pretrained, MASTER_64_URL, hidden=64)


def add_model_flags(parser):
    group = parser.add_mutually_exclusive_group(required=False)
    group.add_argument(
        "-m", "--model_path", help="Path to local trained model."
    )
    group.add_argument(
        "--dns48", action="store_true",
        help="Use pre-trained real time H=48 model trained on DNS."
    )
    group.add_argument(
        "--dns64", action="store_true",
        help="Use pre-trained real time H=64 model trained on DNS."
    )
    group.add_argument(
        "--master64", action="store_true",
        help="Use pre-trained real time H=64 model trained on DNS and Valentini."
    )


def get_model(args):
    """
    Load local model package or torchhub pre-trained model.
    """
    if args.model_path:
        logger.info("Loading model from %s", args.model_path)
        pkg = torch.load(args.model_path)
        model = deserialize_model(pkg)
    elif args.dns64:
        logger.info("Loading pre-trained real time H=64 model trained on DNS.")
        model = dns64()
    elif args.master64:
        logger.info(
            "Loading pre-trained real time H=64 model trained on DNS and Valentini."
        )
        model = master64()
    else:
        logger.info("Loading pre-trained real time H=48 model trained on DNS.")
        model = dns48()
    logger.debug(model)
    return model



from .code2voice import Code2Speech
from .interface import HFSpeechInference, HFWhisperInference, WhisperInference
from .live import LiveSpeech, live_asr_process, live_vad_process, live_list_microphones, live_get_input_device_id
from .metric import chunked_wer, chunked_cer, wer, cer
from .preprocessing import *
from .seeak import Seeak
from .speaker_embedding import extract_d_vector, extract_x_vector
from .speech_enhancement import SpeechEnhancer
from .voice2code import HubertCode
from .voice2code_model import *


# Copyright (c) Facebook, Inc. and its affiliates.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.
# author: adefossez

import math
import time

import torch as th
from torch import nn
from torch.nn import functional as F

from .resample import downsample2, upsample2
from .utils import capture_init


class BLSTM(nn.Module):
    def __init__(self, dim, layers=2, bi=True):
        super().__init__()
        klass = nn.LSTM
        self.lstm = klass(
            bidirectional=bi, num_layers=layers, hidden_size=dim, input_size=dim
        )
        self.linear = None
        if bi:
            self.linear = nn.Linear(2 * dim, dim)

    def forward(self, x, hidden=None):
        x, hidden = self.lstm(x, hidden)
        if self.linear:
            x = self.linear(x)
        return x, hidden


def rescale_conv(conv, reference):
    std = conv.weight.std().detach()
    scale = (std / reference)**0.5
    conv.weight.data /= scale
    if conv.bias is not None:
        conv.bias.data /= scale


def rescale_module(module, reference):
    for sub in module.modules():
        if isinstance(sub, (nn.Conv1d, nn.ConvTranspose1d)):
            rescale_conv(sub, reference)


class Demucs(nn.Module):
    """
    Demucs speech enhancement model.
    Args:
        - chin (int): number of input channels.
        - chout (int): number of output channels.
        - hidden (int): number of initial hidden channels.
        - depth (int): number of layers.
        - kernel_size (int): kernel size for each layer.
        - stride (int): stride for each layer.
        - causal (bool): if false, uses BiLSTM instead of LSTM.
        - resample (int): amount of resampling to apply to the input/output.
            Can be one of 1, 2 or 4.
        - growth (float): number of channels is multiplied by this for every layer.
        - max_hidden (int): maximum number of channels. Can be useful to
            control the size/speed of the model.
        - normalize (bool): if true, normalize the input.
        - glu (bool): if true uses GLU instead of ReLU in 1x1 convolutions.
        - rescale (float): controls custom weight initialization.
            See https://arxiv.org/abs/1911.13254.
        - floor (float): stability flooring when normalizing.

    """
    @capture_init
    def __init__(self,
                 chin=1,
                 chout=1,
                 hidden=48,
                 depth=5,
                 kernel_size=8,
                 stride=4,
                 causal=True,
                 resample=4,
                 growth=2,
                 max_hidden=10_000,
                 normalize=True,
                 glu=True,
                 rescale=0.1,
                 floor=1e-3):

        super().__init__()
        if resample not in [1, 2, 4]:
            raise ValueError("Resample should be 1, 2 or 4.")

        self.chin = chin
        self.chout = chout
        self.hidden = hidden
        self.depth = depth
        self.kernel_size = kernel_size
        self.stride = stride
        self.causal = causal
        self.floor = floor
        self.resample = resample
        self.normalize = normalize

        self.encoder = nn.ModuleList()
        self.decoder = nn.ModuleList()
        activation = nn.GLU(1) if glu else nn.ReLU()
        ch_scale = 2 if glu else 1

        for index in range(depth):
            encode = []
            encode += [
                nn.Conv1d(chin, hidden, kernel_size, stride),
                nn.ReLU(),
                nn.Conv1d(hidden, hidden * ch_scale, 1), activation,
            ]
            self.encoder.append(nn.Sequential(*encode))

            decode = []
            decode += [
                nn.Conv1d(hidden, ch_scale * hidden, 1), activation,
                nn.ConvTranspose1d(hidden, chout, kernel_size, stride),
            ]
            if index > 0:
                decode.append(nn.ReLU())
            self.decoder.insert(0, nn.Sequential(*decode))
            chout = hidden
            chin = hidden
            hidden = min(int(growth * hidden), max_hidden)

        self.lstm = BLSTM(chin, bi=not causal)
        if rescale:
            rescale_module(self, reference=rescale)

    def valid_length(self, length):
        """
        Return the nearest valid length to use with the model so that
        there is no time steps left over in a convolutions, e.g. for all
        layers, size of the input - kernel_size % stride = 0.

        If the mixture has a valid length, the estimated sources
        will have exactly the same length.
        """
        length = math.ceil(length * self.resample)
        for _ in range(self.depth):
            length = math.ceil((length - self.kernel_size) / self.stride) + 1
            length = max(length, 1)
        for _ in range(self.depth):
            length = (length - 1) * self.stride + self.kernel_size
        length = int(math.ceil(length / self.resample))
        return int(length)

    @property
    def total_stride(self):
        return self.stride ** self.depth // self.resample

    def forward(self, mix):
        if mix.dim() == 2:
            mix = mix.unsqueeze(1)

        if self.normalize:
            mono = mix.mean(dim=1, keepdim=True)
            std = mono.std(dim=-1, keepdim=True)
            mix = mix / (self.floor + std)
        else:
            std = 1
        length = mix.shape[-1]
        x = mix
        x = F.pad(x, (0, self.valid_length(length) - length))
        if self.resample == 2:
            x = upsample2(x)
        elif self.resample == 4:
            x = upsample2(x)
            x = upsample2(x)
        skips = []
        for encode in self.encoder:
            x = encode(x)
            skips.append(x)
        x = x.permute(2, 0, 1)
        x, _ = self.lstm(x)
        x = x.permute(1, 2, 0)
        for decode in self.decoder:
            skip = skips.pop(-1)
            x = x + skip[..., :x.shape[-1]]
            x = decode(x)
        if self.resample == 2:
            x = downsample2(x)
        elif self.resample == 4:
            x = downsample2(x)
            x = downsample2(x)

        x = x[..., :length]
        return std * x


def fast_conv(conv, x):
    """
    Faster convolution evaluation if either kernel size is 1
    or length of sequence is 1.
    """
    batch, chin, length = x.shape
    chout, chin, kernel = conv.weight.shape
    assert batch == 1
    if kernel == 1:
        x = x.view(chin, length)
        out = th.addmm(conv.bias.view(-1, 1),
                       conv.weight.view(chout, chin), x)
    elif length == kernel:
        x = x.view(chin * kernel, 1)
        out = th.addmm(conv.bias.view(-1, 1),
                       conv.weight.view(chout, chin * kernel), x)
    else:
        out = conv(x)
    return out.view(batch, chout, -1)


class DemucsStreamer:
    """
    Streaming implementation for Demucs. It supports being fed with any amount
    of audio at a time. You will get back as much audio as possible at that
    point.

    Args:
        - demucs (Demucs): Demucs model.
        - dry (float): amount of dry (e.g. input) signal to keep. 0 is maximum
            noise removal, 1 just returns the input signal. Small values > 0
            allows to limit distortions.
        - num_frames (int): number of frames to process at once. Higher values
            will increase overall latency but improve the real time factor.
        - resample_lookahead (int): extra lookahead used for the resampling.
        - resample_buffer (int): size of the buffer of previous inputs/outputs
            kept for resampling.
    """
    def __init__(self, demucs,
                 dry=0,
                 num_frames=1,
                 resample_lookahead=64,
                 resample_buffer=256):
        device = next(iter(demucs.parameters())).device
        self.demucs = demucs
        self.lstm_state = None
        self.conv_state = None
        self.dry = dry
        self.resample_lookahead = resample_lookahead
        resample_buffer = min(demucs.total_stride, resample_buffer)
        self.resample_buffer = resample_buffer
        self.frame_length = demucs.valid_length(1) + \
            demucs.total_stride * (num_frames - 1)
        self.total_length = self.frame_length + self.resample_lookahead
        self.stride = demucs.total_stride * num_frames
        self.resample_in = th.zeros(demucs.chin, resample_buffer, device=device)
        self.resample_out = th.zeros(
            demucs.chin, resample_buffer, device=device
        )

        self.frames = 0
        self.total_time = 0
        self.variance = 0
        self.pending = th.zeros(demucs.chin, 0, device=device)

        bias = demucs.decoder[0][2].bias
        weight = demucs.decoder[0][2].weight
        chin, chout, kernel = weight.shape
        self._bias = bias.view(-1, 1).repeat(1, kernel).view(-1, 1)
        self._weight = weight.permute(1, 2, 0).contiguous()

    def reset_time_per_frame(self):
        self.total_time = 0
        self.frames = 0

    @property
    def time_per_frame(self):
        return self.total_time / self.frames

    def flush(self):
        """
        Flush remaining audio by padding it with zero. Call this
        when you have no more input and want to get back the last chunk of audio.
        """
        pending_length = self.pending.shape[1]
        padding = th.zeros(
            self.demucs.chin, self.total_length, device=self.pending.device
        )
        out = self.feed(padding)
        return out[:, :pending_length]

    def feed(self, wav):
        """
        Apply the model to mix using true real time evaluation.
        Normalization is done online as is the resampling.
        """
        begin = time.time()
        demucs = self.demucs
        resample_buffer = self.resample_buffer
        stride = self.stride
        resample = demucs.resample

        if wav.dim() != 2:
            raise ValueError("input wav should be two dimensional.")
        chin, _ = wav.shape
        if chin != demucs.chin:
            raise ValueError(f"Expected {demucs.chin} channels, got {chin}")

        self.pending = th.cat([self.pending, wav], dim=1)
        outs = []
        while self.pending.shape[1] >= self.total_length:
            self.frames += 1
            frame = self.pending[:, :self.total_length]
            dry_signal = frame[:, :stride]
            if demucs.normalize:
                mono = frame.mean(0)
                variance = (mono**2).mean()
                self.variance = variance / self.frames + \
                    (1 - 1 / self.frames) * self.variance
                frame = frame / (demucs.floor + math.sqrt(self.variance))
            frame = th.cat([self.resample_in, frame], dim=-1)
            self.resample_in[:] = frame[:, stride - resample_buffer:stride]

            if resample == 4:
                frame = upsample2(upsample2(frame))
            elif resample == 2:
                frame = upsample2(frame)
            # remove pre sampling buffer
            frame = frame[:, resample * resample_buffer:]
            # remove extra samples after window
            frame = frame[:, :resample * self.frame_length]

            out, extra = self._separate_frame(frame)
            padded_out = th.cat([self.resample_out, out, extra], 1)
            self.resample_out[:] = out[:, -resample_buffer:]
            if resample == 4:
                out = downsample2(downsample2(padded_out))
            elif resample == 2:
                out = downsample2(padded_out)
            else:
                out = padded_out

            out = out[:, resample_buffer // resample:]
            out = out[:, :stride]

            if demucs.normalize:
                out *= math.sqrt(self.variance)
            out = self.dry * dry_signal + (1 - self.dry) * out
            outs.append(out)
            self.pending = self.pending[:, stride:]

        self.total_time += time.time() - begin
        if outs:
            out = th.cat(outs, 1)
        else:
            out = th.zeros(chin, 0, device=wav.device)
        return out

    def _separate_frame(self, frame):
        demucs = self.demucs
        skips = []
        next_state = []
        first = self.conv_state is None
        stride = self.stride * demucs.resample
        x = frame[None]
        for idx, encode in enumerate(demucs.encoder):
            stride //= demucs.stride
            length = x.shape[2]
            if idx == demucs.depth - 1:
                # This is sligthly faster for the last conv
                x = fast_conv(encode[0], x)
                x = encode[1](x)
                x = fast_conv(encode[2], x)
                x = encode[3](x)
            else:
                if not first:
                    prev = self.conv_state.pop(0)
                    prev = prev[..., stride:]
                    tgt = (length - demucs.kernel_size) // demucs.stride + 1
                    missing = tgt - prev.shape[-1]
                    offset = length - demucs.kernel_size - \
                        demucs.stride * (missing - 1)
                    x = x[..., offset:]
                x = encode[1](encode[0](x))
                x = fast_conv(encode[2], x)
                x = encode[3](x)
                if not first:
                    x = th.cat([prev, x], -1)
                next_state.append(x)
            skips.append(x)

        x = x.permute(2, 0, 1)
        x, self.lstm_state = demucs.lstm(x, self.lstm_state)
        x = x.permute(1, 2, 0)
        # In the following, x contains only correct samples, i.e. the one
        # for which each time position is covered by two window of the upper
        # layer. extra contains extra samples to the right, and is used only as
        # a better padding for the online resampling.
        extra = None
        for idx, decode in enumerate(demucs.decoder):
            skip = skips.pop(-1)
            x += skip[..., :x.shape[-1]]
            x = fast_conv(decode[0], x)
            x = decode[1](x)

            if extra is not None:
                skip = skip[..., x.shape[-1]:]
                extra += skip[..., :extra.shape[-1]]
                extra = decode[2](decode[1](decode[0](extra)))
            x = decode[2](x)
            next_state.append(
                x[..., -demucs.stride:] - decode[2].bias.view(-1, 1)
            )
            if extra is None:
                extra = x[..., -demucs.stride:]
            else:
                extra[..., :demucs.stride] += next_state[-1]
            x = x[..., :-demucs.stride]

            if not first:
                prev = self.conv_state.pop(0)
                x[..., :demucs.stride] += prev
            if idx != demucs.depth - 1:
                x = decode[3](x)
                extra = decode[3](extra)
        self.conv_state = next_state
        return x[0], extra[0]


def test():
    import argparse
    parser = argparse.ArgumentParser(
        "denoiser.demucs",
        description="Benchmark the streaming Demucs implementation, as well as "
                    "checking the delta with the offline implementation.")
    parser.add_argument("--depth", default=5, type=int)
    parser.add_argument("--resample", default=4, type=int)
    parser.add_argument("--hidden", default=48, type=int)
    parser.add_argument("--sample_rate", default=16000, type=float)
    parser.add_argument("--device", default="cpu")
    parser.add_argument("-t", "--num_threads", type=int)
    parser.add_argument("-f", "--num_frames", type=int, default=1)
    args = parser.parse_args()
    if args.num_threads:
        th.set_num_threads(args.num_threads)
    sr = args.sample_rate
    sr_ms = sr / 1000
    demucs = Demucs(
        depth=args.depth, hidden=args.hidden, resample=args.resample
    ).to(args.device)
    x = th.randn(1, int(sr * 4)).to(args.device)
    out = demucs(x[None])[0]
    streamer = DemucsStreamer(demucs, num_frames=args.num_frames)
    out_rt = []
    frame_size = streamer.total_length
    with th.no_grad():
        while x.shape[1] > 0:
            out_rt.append(streamer.feed(x[:, :frame_size]))
            x = x[:, frame_size:]
            frame_size = streamer.demucs.total_stride
    out_rt.append(streamer.flush())
    out_rt = th.cat(out_rt, 1)
    model_size = sum(p.numel() for p in demucs.parameters()) * 4 / 2**20
    initial_lag = streamer.total_length / sr_ms
    tpf = 1000 * streamer.time_per_frame
    print(f"model size: {model_size:.1f}MB, ", end='')
    print(f"delta batch/streaming: {th.norm(out - out_rt) / th.norm(out):.2%}")
    print(f"initial lag: {initial_lag:.1f}ms, ", end='')
    print(f"stride: {streamer.stride * args.num_frames / sr_ms:.1f}ms")
    print(f"time per frame: {tpf:.1f}ms, ", end='')
    rtf = (1000 * streamer.time_per_frame) / (streamer.stride / sr_ms)
    print(f"RTF: {rtf:.2f}")
    print(f"Total lag with computation: {initial_lag + tpf:.1f}ms")


if __name__ == "__main__":
    test()



import random

import torch
import torchaudio


def cal_audio_time(filename):
    import subprocess, json
    result = subprocess.check_output(
        f'ffprobe -v quiet -show_format -of json "{filename}"',
        shell=True).decode()
    duration = json.loads(result)['format']['duration']
    return float(duration)


def shuffle_gen(n):
    # this is used like a range(n) list, but we don’t store
    # those entries where state[i] = i.
    state = dict()
    for remaining in range(n, 0, -1):
        i = random.randrange(remaining)
        yield state.get(i, i)
        state[i] = state.get(remaining - 1, remaining - 1)
        # Cleanup – we don’t need this information anymore
        state.pop(remaining - 1, None)


def write(wav, filename, sr=16_000):
    # Normalize audio if it prevents clipping
    wav = wav / max(wav.abs().max().item(), 1)
    torchaudio.save(filename, wav.cpu(), sr)


def collate_fn_padd(batch, device):
    '''
    Padds batch of variable length

    note: it converts things ToTensor manually here since the ToTensor transform
    assume it takes in images rather than arbitrary tensors.
    '''
    ## get sequence lengths
    lengths = torch.tensor([t.shape[0] for t in batch]).to(device)
    ## padd
    batch = [torch.Tensor(t).to(device) for t in batch]
    batch = torch.nn.utils.rnn.pad_sequence(batch, batch_first=True, padding_value=-100)
    ## compute mask
    mask = (batch != -100).to(device)
    return batch, lengths, mask



# Copyright (c) Facebook, Inc. and its affiliates.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.
# author: adefossez

import functools
import logging
from contextlib import contextmanager
import inspect
import time

logger = logging.getLogger(__name__)

EPS = 1e-8


def capture_init(init):
    """capture_init.

    Decorate `__init__` with this, and you can then
    recover the *args and **kwargs passed to it in `self._init_args_kwargs`
    """
    @functools.wraps(init)
    def __init__(self, *args, **kwargs):
        self._init_args_kwargs = (args, kwargs)
        init(self, *args, **kwargs)

    return __init__


def deserialize_model(package, strict=False):
    """deserialize_model.

    """
    klass = package['class']
    if strict:
        model = klass(*package['args'], **package['kwargs'])
    else:
        sig = inspect.signature(klass)
        kw = package['kwargs']
        for key in list(kw):
            if key not in sig.parameters:
                logger.warning("Dropping inexistant parameter %s", key)
                del kw[key]
        model = klass(*package['args'], **kw)
    model.load_state_dict(package['state'])
    return model


def copy_state(state):
    return {k: v.cpu().clone() for k, v in state.items()}


def serialize_model(model):
    args, kwargs = model._init_args_kwargs
    state = copy_state(model.state_dict())
    return {"class": model.__class__, "args": args, "kwargs": kwargs, "state": state}


@contextmanager
def swap_state(model, state):
    """
    Context manager that swaps the state of a model, e.g:

        # model is in old state
        with swap_state(model, new_state):
            # model in new state
        # model back to old state
    """
    old_state = copy_state(model.state_dict())
    model.load_state_dict(state)
    try:
        yield
    finally:
        model.load_state_dict(old_state)


def pull_metric(history, name):
    out = []
    for metrics in history:
        if name in metrics:
            out.append(metrics[name])
    return out


class LogProgress:
    """
    Sort of like tqdm but using log lines and not as real time.
    Args:
        - logger: logger obtained from `logging.getLogger`,
        - iterable: iterable object to wrap
        - updates (int): number of lines that will be printed, e.g.
            if `updates=5`, log every 1/5th of the total length.
        - total (int): length of the iterable, in case it does not support
            `len`.
        - name (str): prefix to use in the log.
        - level: logging level (like `logging.INFO`).
    """
    def __init__(self,
                 logger,
                 iterable,
                 updates=5,
                 total=None,
                 name="LogProgress",
                 level=logging.INFO):
        self.iterable = iterable
        self.total = total or len(iterable)
        self.updates = updates
        self.name = name
        self.logger = logger
        self.level = level

    def update(self, **infos):
        self._infos = infos

    def __iter__(self):
        self._iterator = iter(self.iterable)
        self._index = -1
        self._infos = {}
        self._begin = time.time()
        return self

    def __next__(self):
        self._index += 1
        try:
            value = next(self._iterator)
        except StopIteration:
            raise
        else:
            return value
        finally:
            log_every = max(1, self.total // self.updates)
            # logging is delayed by 1 it, in order to have the metrics from update
            if self._index >= 1 and self._index % log_every == 0:
                self._log()

    def _log(self):
        self._speed = (1 + self._index) / (time.time() - self._begin)
        infos = " | ".join(f"{k.capitalize()} {v}" for k, v in self._infos.items())
        if self._speed < 1e-4:
            speed = "oo sec/it"
        elif self._speed < 0.1:
            speed = f"{1/self._speed:.1f} sec/it"
        else:
            speed = f"{self._speed:.1f} it/sec"
        out = f"{self.name} | {self._index}/{self.total} | {speed}"
        if infos:
            out += " | " + infos
        self.logger.log(self.level, out)


def colorize(text, color):
    """
    Display text with some ANSI color in the terminal.
    """
    code = f"\033[{color}m"
    restore = "\033[0m"
    return "".join([code, text, restore])


def bold(text):
    """
    Display text in bold in the terminal.
    """
    return colorize(text, "1")


def cal_snr(lbl, est):
    import torch
    y = 10.0 * torch.log10(
        torch.sum(lbl**2, dim=-1) / (torch.sum((est-lbl)**2, dim=-1) + EPS) +
        EPS
    )
    return y



import os

import nlp2

from asrp import HubertCode


def hubert_layer9_code500(sampling_rate=16000,
                          chunk_sec=10,
                          worker=8,
                          return_diff=False,
                          batch=None):
    nlp2.download_file(
        'https://dl.fbaipublicfiles.com/hubert/hubert_base_ls960_L9_km500.bin', './')
    hc = HubertCode("facebook/hubert-base-ls960", './hubert_base_ls960_L9_km500.bin', 9,
                    sampling_rate=sampling_rate,
                    chunk_sec=chunk_sec,
                    worker=worker,
                    return_diff=return_diff,
                    batch=batch)
    return hc


def hubert_layer6_code50(sampling_rate=16000,
                         chunk_sec=10,
                         worker=8,
                         return_diff=False,
                         batch=None):
    nlp2.download_file(
        'https://dl.fbaipublicfiles.com/textless_nlp/gslm/hubert/km50/km.bin', './')
    os.rename('./km.bin', './hubert_base_ls960_L6_km50.bin')
    hc = HubertCode("facebook/hubert-base-ls960", './hubert_base_ls960_L6_km50.bin', 6,
                    sampling_rate=sampling_rate,
                    chunk_sec=chunk_sec,
                    worker=worker,
                    return_diff=return_diff,
                    batch=batch)
    return hc


def hubert_layer6_code100(sampling_rate=16000,
                          chunk_sec=10,
                          worker=8,
                          return_diff=False,
                          batch=None):
    nlp2.download_file(
        'https://dl.fbaipublicfiles.com/textless_nlp/gslm/hubert/km100/km.bin', './')
    os.rename('./km.bin', './hubert_base_ls960_L6_km100.bin')
    hc = HubertCode("facebook/hubert-base-ls960", './hubert_base_ls960_L6_km100.bin', 6,
                    sampling_rate=sampling_rate,
                    chunk_sec=chunk_sec,
                    worker=worker,
                    return_diff=return_diff,
                    batch=batch)
    return hc


def hubert_layer6_code200(sampling_rate=16000,
                          chunk_sec=10,
                          worker=8,
                          return_diff=False,
                          batch=None):
    nlp2.download_file(
        'https://dl.fbaipublicfiles.com/textless_nlp/gslm/hubert/km200/km.bin', './')
    os.rename('./km.bin', './hubert_base_ls960_L6_km200.bin')
    hc = HubertCode("facebook/hubert-base-ls960", './hubert_base_ls960_L6_km200.bin', 6,
                    sampling_rate=sampling_rate,
                    chunk_sec=chunk_sec,
                    worker=worker,
                    return_diff=return_diff,
                    batch=batch)
    return hc
